""" **Description**
        Initialize.

    **License**
        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-03-22.
"""

from .ComplexTransform import ComplexTransform
from .FourierTransform import FourierTransform
from .PowerSpectrumTransform import PowerSpectrumTransform
from .WaveletTransform import WaveletTransform
from .ZTransform import ZTransform

__all__ = [ 'ComplexTransform', 'FourierTransform', 'PowerSpectrumTransform', 'WaveletTransform',
            'ZTransform' ]



""" **Description**
        A power spectrum transform converts a real or complex discrete-time
        incident signal to a real discrete-frequency reference signal, which
        estimates a mean power density of the incident signal relative to
        frequency.  A forward coefficient array is specified to define a
        window filter.

        Singleton.

        A power spectrum transform is constructed by estimating a mean power
        from a collection of Fourier transforms of an incident signal, over a
        sliding window defined by a forward coefficient array which defines a
        window filter.  An index specifies a sample interval, or a
        non-overlapping stride, between successive operations.

        .. math::
            v_{i,k} = \\frac{1}{N}\ \sum_{n = 0}^{N-1} b_{n} x_{n+i\ I} e^{ \\frac{\ -j\ \pi\ k \ n}{N} }

        .. math::
            y_{k} = \\frac{1}{C}\ \sum_{i = 0}^{C-1} v_{i,k} v^{*}_{i,k}

        A spectrogram may be electively defined such that the collection of
        Fourier transforms is preserved to construct a time frequency
        representation of the power spectrum.

        A power spectrum transform is normalized by incident signal length and
        forms a contiguous sequence corresponding to a linear and increasing
        normalized frequency.

        .. math::
            f_{k} = \ 2\ \\frac{k}{N}

        An incident signal length is inversely proportional to a normalized
        frequency resolution.

        .. math::
            N = \\frac{2}{R}

    **Example**
       
        ::
        
            from diamondback import ComplexExponentialFilter, PowerSpectrumTransform
            import numpy

            x = ComplexExponentialFilter( 0.0 ).filter( numpy.linspace( 0.12, 0.23, 1024 ) ) * numpy.random.rand( 1 )[ 0 ]
            b = WindowFilter( 'Hann', 128 - 1 ).b

            # Transform an incident signal.

            y, f = PowerSpectrumTransform.transform( x, b = b, index = len( b ) // 2 )

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-04-13.
"""

from diamondback.transforms.FourierTransform import FourierTransform
from typing import List, Tuple, Union
import numpy

class PowerSpectrumTransform( object ) :

    """ Power spectrum transform.
    """

    @staticmethod
    def transform( x : Union[ List, numpy.ndarray ], b : Union[ List, numpy.ndarray ], index : int, spectrogram : bool = False ) -> Tuple[ numpy.ndarray, numpy.ndarray ] :

        """ Transforms a real or complex discrete-time incident signal to a
            real discrete-frequency reference signal.

            Arguments :
                x : Union[ List, numpy.ndarray ] - incident signal.
                b : Union[ List, numpy.ndarray ] - forward coefficient.
                index : int.
                spectrogram : bool.

            Returns :
                y : numpy.ndarray - reference signal.
                f : numpy.ndarray - frequency normalized to Nyquist in [ -1.0, 1.0 ).
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :
            x = numpy.array( list( x ) )
        if ( not len( x ) ) :
            raise ValueError( f'X = {x}' )
        if ( ( not numpy.isscalar( b ) ) and ( not isinstance( b, numpy.ndarray ) ) ) :
            b = numpy.array( list( b ) )
        if ( ( not len( b ) ) or ( numpy.isclose( b, 0.0 ).all( ) ) ) :
            raise ValueError( f'B = {b}' )
        if ( len( x ) < len( b ) ) :
            raise ValueError( f'X = {x}' )
        y, f = [ ], None
        jj = len( b ) // 2
        for ii in range( 0, len( x ) - len( b ) + 1, index ) :
            v, f = FourierTransform.transform( x[ ii : ii + len( b ) ], b )
            y.append( abs( v[ jj : ] ) ** 2 )
        y, f = numpy.stack( y ) if ( spectrogram ) else numpy.sum( y, axis = 0 ) / len( y ), f[ jj : ]
        return y, f



""" **Description**
        Test unit commons.
    
    **Example**
      
        ::
        
            pytest --capture=no --verbose
    
    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.
    
    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-04-03.
"""

from diamondback import IirFilter
from diamondback import Log, RestClient, Serial
import io
import numpy
import pytest
import sys

class Test( object ) :

    """ Test.
    """

    def test_Log( self ) :

        """ Test Log.
        """

        x = [ str( x ) for x in numpy.random.rand( 3 ) ]
        fio = io.StringIO( )
        Log.stream( fio )
        Log.level( 'CRITICAL' )
        Log.write( 'DEBUG', f'{numpy.random.rand( 3 ) + 1.0}' )
        Log.level( 'DEBUG' )
        Log.write( 'INFO', x[ 0 ] )
        Log.level( 'ERROR' )
        Log.write( 'WARNING', f'{numpy.random.rand( 3 ) + 1.0}' )
        Log.write( 'CRITICAL', x[ 1 ] )
        Log.write( 'ERROR', x[ 2 ] )
        value = fio.getvalue( )
        Log.stream( sys.stdout )
        Log.level( 'INFO' )
        assert value
        assert value.find( x[ 0 ] ) > 0
        assert value.find( x[ 1 ] ) > value.find( x[ 0 ] )
        assert value.find( x[ 2 ] ) > value.find( x[ 1 ] )

    def test_RestClient( self ) :

        """ Test RestClient.
        """

        client = RestClient( )
        client.url = 'http://en.wikipedia.org'
        assert client.live
        assert client.request( 'get','/wiki/Marines' ).content

    def test_Serial( self ) :

        """ Test Serial.
        """

        x = IirFilter( style = 'Butterworth', frequency = 0.1, order = 4, count = 1 )
        for ii in range( 0, 2 ) :
            y = Serial.decode( Serial.encode( x, ii != 0 ), ii != 0 )
            assert ( ( numpy.allclose( x.a, y.a ) ) and ( numpy.allclose( x.b, y.b ) ) and ( numpy.allclose( x.s, y.s ) ) )
            assert Serial.code( Serial.encode( x, compress = False ) ) == Serial.code( Serial.encode( y, compress = False ) )
            with pytest.raises( ValueError ) :
                Serial.decode( Serial.encode( x, bool( ii ) ), not ii )
        x = dict( x = numpy.random.rand( 30, 50 ), y = numpy.random.rand( 50, 30 ) )
        for ii in range( 0, 2 ) :
            y = Serial.decode( Serial.encode( x, ii != 0 ), ii != 0 )
            assert Serial.encode( x, compress = False ) == Serial.encode( y, compress = False )
            assert all( [ u in x for u in list( y.keys( ) ) ] )
        x = '\"\"\" Docstring. \"\"\"\n \
                    { "serial" : { "datetime" : "2019-09-01T15:10:00Z", # Comment.\n \
                                   "duration" : 28800.0,\n \
                                   "period" : 123.4 } }\n'
        y = Serial.decode( x, compress = False, clean = True )
        assert 'serial' in y
        assert all( [ u in y[ 'serial' ] for u in ( 'datetime', 'duration', 'period' ) ] )
        assert numpy.isclose( y[ 'serial' ][ 'period' ], 123.4 )



""" **Description**
        An Infinite Impulse Response ( IIR ) filter realizes a discrete
        difference equation as a function of a recursive coefficient array,
        a forward coefficient array, and a state array of a specified order,
        consuming an incident signal and producing a reference signal.

        .. math::
            y_{n} = \sum_{i = 1}^{N} a_{i,n} y_{n-i} + \sum_{i = 0}^{N} b_{i,n} x_{n-i} = \sum_{i = 1}^{N} (\ a_{i,n} b_{0,n} + b_{i,n}\ ) s_{i,n} + b_{0,n} x_{n}\qquad a_{0,n} = 0

        .. math::
            s_{1,n+1} = \sum_{i = 1}^{N} a_{i,n} s_{i,n} + x_{n}\qquad\qquad s_{i,n+1} = s_{i-1,n}

        A reset may minimize edge effects at a discontinuity by assuming
        persistent operation at a specified incident signal condition.

        .. math::
            s_{i,n} = \\frac{1.0 - b_{0,n}}{\sum_{i=1}^{N} a_{i,n} b_{0,n} + b_{i,n}}\ x_{n}

        A frequency response is expressed as a function of a recursive
        coefficient array and a forward coefficient array.

        .. math::
            H_{z,n} = \\frac{\sum_{i = 0}^{N} b_{i,n} z^{-i}}{{1 - \sum_{i = 1}^{N} a_{i,n} z^{-i}}}

        A recursive coefficient array, forward coefficient array,
        and state array of a specified order are defined to realize specified
        constraints.  A style, frequency, order, count, complement,
        and gain are electively specified.  Alternatively, a recursive coefficient
        array, forward coefficient array, and state array may be explicitly
        defined to ignore constraints.

        Frequency corresponds to a -3 dB frequency response normalized relative
        to Nyquist.

        Style is in ( 'Bessel', 'Butterworth', 'Chebyshev' ).

        * | 'Bessel' filters demonstrate maximally linear phase response or
          | constant group delay.

        * | 'Butterworth' filters demonstrate maximally flat magnitude response.

        * | 'Chebyshev' filters demonstrate minimally low magnitude response error
          | and improved rate of attenuation in a Type I form, with a maximum in
          | band ripple of 0.125 dB.

        Count is a quantity of filters of a specified order concatenated to
        form an aggregate frequency response in cascade form.

        Complement effectively constructs a mirror image of a specified
        frequency response.

    **Example**
       
        ::
        
            from diamondback import IirFilter
            import numpy

            # Create an instance with constraints.

            obj = IirFilter( style = 'Chebyshev', frequency = 0.1, order = 8, count = 1 )

            # Create an instance with coefficients.

            obj = IirFilter( a = obj.a, b = obj.b )

            # Estimate frequency response, group delay, and roots.

            y, f = obj.response( length = 8192, count = 1 )
            y, f = obj.delay( length = 8192, count = 1 )
            p, z = obj.roots( count = 1 )

            # Filter an incident signal.

            x = numpy.random.rand( 128 ) * 2.0 - 1.0
            obj.reset( x[ 0 ] )
            y = obj.filter( x )

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-01-23.
"""

from diamondback.filters.FirFilter import FirFilter
from diamondback.transforms.ZTransform import ZTransform
from typing import List, Tuple, Union
import math
import numpy
import scipy.signal
import warnings

class IirFilter( FirFilter ) :

    """ Infinite Impulse Response ( IIR ) filter.
    """

    __style = ( 'Bessel', 'Butterworth', 'Chebyshev' )

    @property
    def a( self ) :

        """ a : Union[ List, numpy.ndarray ] - recursive coefficient.
        """

        return self._a

    @a.setter
    def a( self, a : Union[ List, numpy.ndarray ] ) :

        self._a = a

    def __init__( self, style : str = '', frequency : float = 0.0, order : int = 0, count : int = 1, complement : bool = False, gain : float = 1.0,
                  a : Union[ List, numpy.ndarray ] = [ ], b : Union[ List, numpy.ndarray ] = [ ], s : Union[ List, numpy.ndarray ] = [ ] ) -> None :

        """ Initialize.

            Specify constraints including style, frequency, and order.
            Alternatively, a recursive coefficient array and forward coefficient array
            may be explicitly defined to ignore constraints.
            
            Labels should be used to avoid ambiguity between constraints and
            coefficients.

            Arguments :
                style : str - in ( 'Bessel', 'Butterworth', 'Chebyshev' ).
                frequency : float - frequency normalized to Nyquist in ( 0.0, 1.0 ).
                order : int - order per instance.
                count : int - instances per cascade.
                complement : bool - complement response.
                gain : float - gain.
                a : Union[ List, numpy.ndarray ] - recursive coefficient.
                b : Union[ List, numpy.ndarray ] - forward coefficient.
                s : Union[ List, numpy.ndarray ] - state.
        """

        if ( ( not len( a ) ) and ( ( not len( b ) ) ) ) :
            if ( ( not style ) or ( style not in IirFilter.__style ) ) :
                raise ValueError( f'style = {style} Expected Style in {IirFilter.__style}' )
            if ( ( frequency <= 0.0 ) or ( frequency >= 1.0 ) ) :
                raise ValueError( f'Frequency = {frequency} Expected Frequency in ( 0.0, 1.0 )' )
            if ( order < 1 ) :
                raise ValueError( f'Order = {order} Expected Order in [ 1, inf )' )
            if ( count < 1 ) :
                raise ValueError( f'Count = {count} Expected Count in [ 1, inf )' )
            if ( complement ) :
                frequency = 1.0 - frequency
            beta, eps, error = 10.0, numpy.finfo( float ).eps, float( 'inf' )
            index, mu, zeta = 500 * ( 1 + ( count > 2 ) ), 2.5e-2, 1.0
            a, b = [ ], [ ]
            for _ in range( 0, index ) :
                u, v = IirFilter._evaluate( style, zeta * frequency, order )
                x = numpy.exp( 1j * math.pi * frequency )
                e = ( 2.0 ** ( -0.5 ) ) - ( ( abs( numpy.polyval( v, x ) / numpy.polyval( numpy.concatenate( ( [ 1.0 ], -u[ 1 : ] ) ), x ) ) ) ** count )
                if ( abs( e ) < error ) :
                    a, b, error = u, v, abs( e )
                    if ( error < ( 10.0 * eps ) ) :
                        break
                zeta = max( zeta + mu * math.tanh( beta * e ), eps )
            if ( complement ) :
                a *= numpy.array( [ ( ( -1.0 ) ** x ) for x in range( 0, len( a ) ) ] )
                b *= numpy.array( [ ( ( -1.0 ) ** x ) for x in range( 0, len( b ) ) ] )
                b /= sum( b * numpy.array( [ ( ( -1.0 ) ** x ) for x in range( 0, len( b ) ) ] ) ) / sum( numpy.concatenate( ( [ 1.0 ], -a[ 1 : ] ) ) * numpy.array( [ ( ( -1.0 ) ** x ) for x in range( 0, len( a ) ) ] ) )
            b *= gain
        if ( ( not numpy.isscalar( a ) ) and ( not isinstance( a, numpy.ndarray ) ) ) :
            a = numpy.array( list( a ) )
        if ( ( len( a ) > 0 ) and ( a[ 0 ] ) ) :
            raise ValueError( f'A = {a}' )
        if ( len( a ) < len( b ) ) :
            a = numpy.concatenate( ( a, numpy.zeros( len( b ) - len( a ) ) ) )
        if ( len( b ) < len( a ) ) :
            b = numpy.concatenate( ( b, numpy.zeros( len( a ) - len( b ) ) ) )
        if ( ( isinstance( a[ 0 ], complex ) ) or ( isinstance( b[ 0 ], complex ) ) ) :
            a, b = numpy.array( a, complex ), numpy.array( b, complex )
        if ( a[ 0 ] != 0.0 ) :
            raise ValueError( f'A = {a}' )
        super( ).__init__( b = b, s = s )
        self._a = numpy.array( a )

    @staticmethod
    def _evaluate( style : str, frequency : float, order : int ) -> Tuple[ numpy.ndarray, numpy.ndarray ] :

        """ Evaluates coefficients.

            Arguments :
                style : str - in ( 'Bessel', 'Butterworth', 'Chebyshev' ).
                frequency : float - frequency normalized to Nyquist in ( 0.0, 1.0 ).
                order : int.

            Returns :
                a : numpy.ndarray - recursive coefficient.
                b : numpy.ndarray - forward coefficient.
        """

        bilinear = True
        if ( style == 'Bessel' ) :
            bilinear = False
            u, a = numpy.ones( 1 ), numpy.ones( 2 )
            for ii in range( 2, order + 1 ) :
                x = numpy.concatenate( ( u, numpy.zeros( 2 ) ) ) + numpy.concatenate( ( [ 0.0 ], ( ( 2.0 * ii ) - 1.0 ) * a ) )
                u, a = a, x
        elif ( style == 'Butterworth' ) :
            a = numpy.ones( 1 )
            for ii in range( 1, ( order // 2 ) + 1 ) :
                a = numpy.convolve( a, numpy.array( [ 1.0, -2.0 * math.cos( ( ( ( 2.0 * ii ) + order - 1.0 ) / ( 2.0 * order ) ) * math.pi ), 1.0 ] ) )
            if ( order & 1 ) :
                a = numpy.convolve( a, numpy.ones( 2 ) )
        elif ( style == 'Chebyshev' ) :
            ripple = 0.125
            u = numpy.array( [ numpy.exp( 1j * math.pi * x / ( 2.0 * order ) ) for x in range( 1, 2 * order, 2 ) ] )
            v = math.asinh( 1.0 / ( ( 10.0 ** ( 0.1 * ripple ) - 1.0 ) ** 0.5 ) ) / order
            a = ( numpy.poly( ( -math.sinh( v ) * u.imag + 1j * math.cosh( v ) * u.real ) * 2.0 * math.pi ) ).real
        a /= a[ -1 ]
        a, b = ZTransform.transform( a, [ 1.0 ], frequency, bilinear )
        b = numpy.poly( -numpy.ones( order ) )
        b *= ( 1.0 - sum( a ) ) / sum( b )
        return a, b

    def delay( self, length : int = 8192, count : int = 1 ) -> Tuple[ numpy.ndarray, numpy.ndarray ] :

        """ Estimates group delay and produces a reference signal.

            Arguments :
                length : int.
                count : int.

            Returns :
                y : numpy.ndarray - reference signal.
                f : numpy.ndarray - frequency normalized to Nyquist in [ -1.0, 1.0 ).
        """

        if ( length < 1 ) :
            raise ValueError( f'Length = {length} Expected Length in [ 1, inf )' )
        if ( count < 1 ) :
            raise ValueError( f'Count = {count} Expected Count in [ 1, inf )' )
        with warnings.catch_warnings( ) :
            warnings.simplefilter( 'ignore' )
            y, f = scipy.signal.group_delay( ( self.b, numpy.concatenate( ( [ 1.0 ], -self.a[ 1 : ] ) ) ), length, True )[ 1 ], numpy.linspace( -1.0, 1.0 - 2.0 / length, length )
            y = numpy.concatenate( ( y[ len( y ) // 2 : ], y[ : len( y ) // 2 ] ) ) * count
        if ( length > 6 ) :
            if ( max( abs( y[ : 6 ] ) ) > min( abs( y[ : 6 ] ) ) * 10.0 ) :
                y[ : 6 ] = numpy.sign( y[ : 6 ] ) * min( abs( y[ : 6 ] ) )
            if ( max( abs( y[ -6 : ] ) ) > min( abs( y[ -6 : ] ) ) * 10.0 ) :
                y[ -6 : ] = numpy.sign( y[ -6 : ] ) * min( abs( y[ -6 : ] ) )
        return y, f

    def filter( self, x : Union[ List, numpy.ndarray ] ) -> numpy.ndarray :

        """ Filters an incident signal and produces a reference signal.

            Arguments :
                x : Union[ List, numpy.ndarray ] - incident signal.

            Returns :
                y : numpy.ndarray - reference signal.
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :
            x = numpy.array( list( x ) )
        if ( not len( x ) ) :
            raise ValueError( f'X = {x}' )
        y = numpy.zeros( len( x ), type( self.b[ 0 ] ) )
        for ii in range( 0, len( x ) ) :
            self.s[ 0 ] = x[ ii ]
            c = self.a * self.b[ 0 ] + self.b
            y[ ii ] = c.dot( self.s )
            if ( len( self.s ) > 1 ) :
                z = self.a.dot( self.s )
                self.s[ 1 : ] = self.s[ : -1 ]
                self.s[ 1 ] += z
        return y

    def reset( self, x : Union[ complex, float ] ) -> None :

        """ Modifies a state to minimize edge effects by assuming persistent
            operation at a specified incident signal condition.

            Arguments :
                x : Union[ complex, float ] - incident signal.
        """

        if ( not numpy.isscalar( x ) ) :
            raise ValueError( f'X = {x}' )
        if ( len( self.s ) > 1 ) :
            self.s.fill( x * ( 1.0 - self.b[ 0 ] ) / ( self.a[ 1 : ] * self.b[ 0 ] + self.b[ 1 : ] ).sum( ) )

    def response( self, length = 8192, count = 1 ) -> Tuple[ numpy.ndarray, numpy.ndarray ] :

        """ Estimates frequency response and produces a reference signal.

            Arguments :
                length : int.
                count : int.

            Returns :
                y : numpy.ndarray - reference signal.
                f : numpy.ndarray - frequency normalized to Nyquist in [ -1.0, 1.0 ).
        """

        if ( length < 1 ) :
            raise ValueError( f'Length = {length} Expected Length in [ 1, inf )' )
        if ( count < 1 ) :
            raise ValueError( f'Count = {count} Expected Count in [ 1, inf )' )
        with warnings.catch_warnings( ) :
            warnings.simplefilter( 'ignore' )
            y, f = scipy.signal.freqz( self.b, numpy.concatenate( ( [ 1.0 ], -self.a[ 1 : ] ) ), length, True )[ 1 ], numpy.linspace( -1.0, 1.0 - 2.0 / length, length )
            y = numpy.concatenate( ( y[ len( y ) // 2 : ], y[ : len( y ) // 2 ] ) ) ** count
        return y, f

    def roots( self, count = 1 ) -> Tuple[ numpy.ndarray, numpy.ndarray ] :

        """ Estimates roots of a frequency response in poles and zeros.

            Arguments :
                count : int.

            Returns :
                p : numpy.ndarray - poles.
                z : numpy.ndarray - zeros.
        """

        p, z = numpy.tile( numpy.roots( numpy.concatenate( ( [ 1.0 ], -self.a[ 1 : ] ) ) ), count ), numpy.tile( numpy.roots( self.b ), count )
        return p[ numpy.argsort( abs( p ) ) ], z[ numpy.argsort( abs( z ) ) ]



""" **Description**
        A Goertzel filter realizes a discrete difference equation which
        approximates a discrete Fourier transform evaluated at a specified
        normalized frequency and order, consuming an incident signal and
        producing a reference signal.

        .. math::
            y_{n} = \sum_{i = 1}^{N} a_{i} y_{n-i} + \sum_{i = 0}^{N} b_{i} x_{n-i} = \sum_{i = 1}^{N} (\ a_{i} b_{0} + b_{i}\ ) s_{i,n} + b_{0} x_{n}\qquad a_{0} = 0

        .. math::
            s_{1,n+1} = \sum_{i = 1}^{N} a_{i} s_{i,n} + x_{n}\qquad\quad s_{i,n+1} = s_{i-1,n}

        .. math::
            \matrix{ a = \scriptsize{ [\ \matrix{ 0 & 2\ \cos(\ \pi\ f\ ) & -1 }\ ] } & b = \scriptsize{ [\ \matrix{ 1 & -e^{\ j\ \pi\ f\ } & 0 } }\ ] }

        At the terminus of each window length a reference signal is evaluated
        to estimate a discrete Fourier transform at a specified normalized
        frequency.

        .. math::
            H_{z} = \\frac{\sum_{i = 0}^{N} b_{i} z^{-i}}{{1 - \sum_{i = 1}^{N} a_{i} z^{-i}}}

        A Goertzel filter is normalized by incident signal length.  An incident
        signal length is is inversely proportional to a normalized frequency
        resolution.

        .. math::
            N = \\frac{2}{R}

    **Example**
       
        ::
        
            from diamondback import ComplexExponentialFilter, GoertzelFilter
            import numpy

            b = WindowFilter( 'Hann', 128 ).b
            frequency = 0.1

            # Create an instance.

            obj = GoertzelFilter( b = b, frequency = frequency )

            # Filter an incident signal.

            x = ComplexExponentialFilter( 0.0 ).filter( numpy.ones( 1024 ) * frequency ) * numpy.random.rand( 1 )[ 0 ]
            y = obj.filter( x )

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-04-16.
"""

from diamondback.filters.IirFilter import IirFilter
from typing import List, Union
import math
import numpy

class GoertzelFilter( IirFilter ) :

    """ Goertzel filter.
    """

    @property
    def frequency( self ) :

        """ frequency : float - frequency normalized to Nyquist in [ -1.0, 1.0 ].
        """

        return self._frequency

    @frequency.setter
    def frequency( self, frequency : float ) :

        if ( ( frequency < -1.0 ) or ( frequency > 1.0 ) ) :
            raise ValueError( f'Frequency = {frequency} Expected Frequency in [ 1.0, 1.0 ]' )
        self._frequency = frequency

    def __init__( self, b : Union[ List, numpy.ndarray ], frequency : float ) -> None :

        """ Initialize.

            Arguments :
                b : Union[ List, numpy.ndarray ] - forward coefficient.
                frequency : float - frequency normalized to Nyquist in [ -1.0, 1.0 ].
        """

        if ( ( not numpy.isscalar( b ) ) and ( not isinstance( b, numpy.ndarray ) ) ) :
            b = numpy.array( list( b ) )
        if ( ( not len( b ) ) or ( numpy.isclose( b, 0.0 ).all( ) ) ) :
            raise ValueError( f'B = {b}' )
        if ( ( frequency < -1.0 ) or ( frequency > 1.0 ) ) :
            raise ValueError( f'Frequency = {frequency} Expected Frequency in [ 1.0, 1.0 ]' )
        u = numpy.array( [ 0.0, 2.0 * math.cos( math.pi * frequency ), -1.0 ] )
        v = numpy.array( [ 1.0, -numpy.exp( -1j * math.pi * frequency ), 0.0 ] )
        super( ).__init__( a = u, b = v )
        self._index, self._w = 0, numpy.array( b )
        self._frequency = frequency

    def filter( self, x : Union[ List, numpy.ndarray ] ) -> numpy.ndarray :

        """ Filters an incident signal and produces a reference signal.

            Arguments :
                x : Union[ List, numpy.ndarray ] - incident signal.

            Returns :
                y : numpy.ndarray - reference signal.
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :
            x = numpy.array( list( x ) )
        if ( not len( x ) ) :
            raise ValueError( f'X = {x}' )
        y = numpy.zeros( int( numpy.ceil( len( x ) / len( self._w ) ) ) + 1, complex )
        u = ( 1.0 + int( not isinstance( x[ 0 ], complex ) ) ) / len( self._w )
        jj = 0
        for ii in range( 0, len( x ) ) :
            v = super( ).filter( u * self._w[ self._index ] * x[ ii : ii + 1 ] )
            self._index += 1
            if ( self._index >= len( self._w ) ) :
                y[ jj ] = v[ 0 ]
                self.s[ : ] = 0.0
                self._index = 0
                jj += 1
        if ( jj != len( y ) ) :
            y = y[ : jj ]
        return y



""" **Description**
        Initialize.

    **License**
        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-03-22.
"""

from .commons import Log, RestClient, Serial
from .filters import FirFilter, IirFilter, ComplexBandPassFilter, ComplexExponentialFilter
from .filters import ComplexFrequencyFilter, DerivativeFilter, GoertzelFilter, IntegralFilter
from .filters import PidFilter, PolynomialRateFilter, PolyphaseRateFilter, RankFilter
from .filters import WindowFilter
from .models import DiversityModel, GaussianModel, GaussianMixtureModel
from .transforms import ComplexTransform, FourierTransform, PowerSpectrumTransform, WaveletTransform
from .transforms import ZTransform

__all__ = [ 'Log', 'RestClient', 'Serial', 'FirFilter',
            'IirFilter', 'ComplexBandPassFilter', 'ComplexExponentialFilter', 'ComplexFrequencyFilter',
            'DerivativeFilter', 'GoertzelFilter', 'IntegralFilter', 'PidFilter',
            'PolynomialRateFilter', 'PolyphaseRateFilter', 'RankFilter', 'WindowFilter',
            'DiversityModel', 'GaussianModel', 'GaussianMixtureModel', 'ComplexTransform',
            'FourierTransform', 'PowerSpectrumTransform', 'WaveletTransform', 'ZTransform' ]
__version__ = '4.1.8'



""" **Description**
        REST client instances define a client for simple REST service requests
        using the requests package.  An API and an elective dictionary of parameter
        strings are encoded to build a URL, elective binary or JSON data are
        defined in the body of a request, and a requests response containing JSON,
        text, or binary data is returned.

        Proxy, timeout, and URL definition are supported.

        Live makes a head request to a URL and detects a live service.

    **Example**
        
        ::
        
            from diamondback import RestClient
            from typing import Dict
            import numpy

            class TestClient( RestClient ) :

                def __init__( self ) -> None :
                    super( ).__init__( )
                    self.proxy = dict( http = '', https = '' )

                def add( self, json : Dict[ str, numpy.ndarray ] ) -> numpy.ndarray :
                    return self.request( 'get', 'test/add', json = json ).json( )

            client = TestClient( )
            client.url = 'http://127.0.0.1:8080'
            client.timeout = ( 10.0, 60.0 )  # connect, read
            value = client.add( dict( x = numpy.random.rand( 3 ), y = numpy.random.rand( 3 ) ) )

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2020-10-22.
"""

from typing import Any, Dict
import requests

class RestClient( object ) :

    """ REST client.
    """

    __method = ( 'delete', 'get', 'head', 'options', 'patch', 'post', 'put' )

    @property
    def live( self ) :

        """ live : bool.
        """

        try :
            requests.request( method = 'head', url = self.url, proxies = self.proxy, timeout = self.timeout )
            value = True
        except Exception :
            value = False
        return value

    @property
    def proxy( self ) :

        """ proxy : Dict[ str, str ].
        """

        return self._proxy

    @proxy.setter
    def proxy( self, proxy : Dict[ str, str ] ) :

        self._proxy = proxy

    @property
    def timeout( self ) :

        """ timeout : Any.
        """

        return self._timeout

    @timeout.setter
    def timeout( self, timeout : Any ) :

        self._timeout = timeout

    @property
    def url( self ) :

        """ url : str.
        """

        return self._url

    @url.setter
    def url( self, url : str ) :

        """ Url.
        """

        if ( url ) :
            url = url.strip( '/' )
        self._url = url

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )
        self._proxy, self._timeout = { }, ( 10.0, 60.0 )
        self._url = 'http://127.0.0.1:8080'

    def request( self, method : str, api : str, auth : Any = None, header : Dict[ str, str ] = None, item : Dict[ str, str ] = None, data : Any = None, json : Any = None ) -> requests.Response :

        """ Request client for simple REST service requests. An API and an
            elective dictionary of parameter strings are encoded to build a
            URL, elective binary or JSON data are defined in the body of a
            request, and a requests response containing JSON, text, or binary
            data is returned.
            
            Arguments :
                method : str - in ( 'delete', 'get', 'head', 'options', 'patch', 'post', 'put' ).
                api : str - relative to the URL.
                auth : Any.
                header : Dict[ str, str ].
                item : Dict[ str, str ].
                data : Any.
                json : Any.

            Returns :
                value : requests.Response.
        """

        if ( not method ) :
            raise ValueError( f'Method = {method} Expected Method in {RestClient.__method}' )
        method = method.lower( )
        if ( method not in RestClient.__method ) :
            raise ValueError( f'Method = {method} Expected Method in {RestClient.__method}' )
        if ( ( data ) and ( json ) ) :
            raise ValueError( f'Data = {data} JSON = {json} Expected Data or JSON' )
        api = api.strip( '/' )
        url = self.url
        if ( api ) :
            url += '/' + api
        with requests.request( method = method, url = url, params = item, data = data, headers = header, auth = auth, json = json, proxies = self.proxy, timeout = self.timeout ) as value :
            value.raise_for_status( )
        return value



""" **Description**
        A Gaussian mixture model is a semi-supervised learning probabilistic
        model instance which uses maximum likelihood estimation, 
        regularization, and expectation maximization to maximize posterior
        probability and classify an incident signal.  Learns model instances
        of a specified order per class, where intra-class models capture
        mixture distributions. 

    **Example**
      
        ::
        
            from diamondback import GaussianMixtureModel

            # Create an instance.

            obj = GaussianMixtureModel( order = 10, index = 100 )

            # Learn an incident signal and predict classification.

            x, y = numpy.random.rand( 32, 2 ), numpy.random.randint( 0, 10, 32 )
            obj.learn( x, y )
            x = numpy.random.rand( 16, 2 )
            v = obj.predict( x )

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-02-08.
"""

import numpy
import sklearn.mixture

class GaussianMixtureModel( object ) :
    
    """ Gaussian mixture model.
    """

    @property
    def index( self ) :
        return self._index

    @index.setter
    def index( self, index ) :
        if ( index < 1 ) :
            raise ValueError( f'Index = {index} Expected Index in [ 1, inf )' )
        self._index = index

    @property
    def order( self ) :
        return self._order

    @property
    def regularize( self ) :
        return self._regularize

    @regularize.setter
    def regularize( self, regularize : float ) :
        if ( regularize < 0.0 ) :
            raise ValueError( f'Regularize = {regularize} Expected Regularize in [ 0.0, inf )' )
        self._regularize = regularize

    @property
    def shape( self ) :
        return self._shape

    def __init__( self, order : int = 10, index : int = 100, regularize : float = 1.0e-1 ) -> None :

        """ Initialize.

            Arguments :
                order : int - mixture distributions per class.
                index : int - iterations.
                regularize : float - regularize.
        """

        if ( order < 1 ) :
            raise ValueError( f'Order = {order} Expected Order in [ 1, inf )' )
        if ( index < 1 ) :
            raise ValueError( f'Index = {index} Expected Index in [ 1, inf )' )
        if ( regularize < 0.0 ) :
            raise ValueError( f'Regularize = {regularize} Expected Regularize in [ 0.0, inf )' )
        self._data = [ ]
        self._index = index
        self._order = order
        self._regularize = regularize
        self._shape = ( )

    def learn( self, x : numpy.ndarray, y : numpy.ndarray ) -> None :

        """ Learns an incident signal with ground truth label and estimates inverse
            covariance and mean matrices to learn mixed distribution instances
            for each class.
        
            Arguments :
                x : numpy.ndarray ( batch, count ) - incident.
                y : numpy.ndarray ( batch ) - label.
        """

        if ( ( len( x.shape ) != 2 ) or ( not all( x.shape ) ) ) :
            raise ValueError( f'X = {x.shape}' )
        if ( len( y ) != x.shape[ 0 ] ) :
            raise ValueError( f'Y = {len( y )} Expected Y = {x.shape[ 0 ]}' )
        if ( not issubclass( y.dtype.type, numpy.integer ) ) :
            raise ValueError( f'Y = {y.dtype.type} Expected Y = {numpy.integer}' )
        self._data = [ ]
        self._shape = x[ 0 ].shape
        for ii in sorted( set( y ) ) :
            z = x[ numpy.where( y == ii )[ 0 ] ]
            m = sklearn.mixture.GaussianMixture( covariance_type = 'full', n_components = self.order, max_iter = self.index, reg_covar = self.regularize )
            m.fit( z )
            self._data.append( m )

    def predict( self, x : numpy.ndarray ) -> numpy.ndarray :

        """ Predicts an estimate of ground truth label from an incident signal
            and maximizes posterior probability of weighted intra-class mixed
            distributions.

            Predictions for each class are ranked and ordered by decending
            probability, and the initial prediction is the most likely class.
        
            Arguments :
                x : numpy.ndarray ( batch, count ) - data.

            Returns :
                v : numpy.ndarray ( batch, class ) - predict.
        """

        if ( not len( self._data ) ) :
            raise ValueError( f'Data = {self._data}' )
        if ( ( len( x.shape ) != 2 ) or ( not all( x.shape ) ) ) :
            raise ValueError( f'X = {x.shape}' )
        if ( x[ 0 ].shape != self._shape ) :
            raise ValueError( f'X = {x[ 0 ].shape} Expected X = {self._shape}' )
        if ( ( not len( self._data ) ) or ( not hasattr( self._data[ 0 ], 'precisions_' ) ) ) :
            raise RuntimeError( f'Model = {self._data} Not Trained' )
        v = numpy.zeros( ( x.shape[ 0 ], len( self._data ) ) )
        for jj in range( 0, len( v ) ) :
            for ii in range( 0, len( self._data ) ) :
                m = self._data[ ii ]
                for kk in range( 0, self.order ) :
                    i, u, w = m.precisions_[ kk ], m.means_[ kk ], m.weights_[ kk ]
                    v[ jj, ii ] += w * numpy.exp( -0.5 * max( ( x[ jj ] - u ) @ i @ ( x[ jj ] - u ).T, 0.0 ) )
        return numpy.fliplr( numpy.argsort( v, axis = 1 ) )



""" **Description**
        A Finite Impulse Response ( FIR ) filter realizes a discrete difference
        equation as a function of a forward coefficient array and a state array
        of a specified order, consuming an incident signal and producing a
        reference signal.

        .. math::
            y_{n} = \sum_{i = 0}^{N} b_{i,n} x_{n-i} = \sum_{i = 1}^{N} b_{i,n} s_{i,n} + b_{0,n} x_{n}

        .. math::
            s_{1,n+1} = x_{n}\qquad\quad s_{i,n+1} = s_{i-1,n}

        A reset may minimize edge effects at a discontinuity by assuming
        persistent operation at a specified incident signal condition.

        .. math::
            s_{i,n} = x_{n}

        A frequency response is expressed as a function of a forward
        coefficient array.

        .. math::
            H_{z,n} = \sum_{i = 0}^{N} b_{i,n} z^{-i}

        A forward coefficient array and state array of a specified order are
        defined to realize specified constraints.  A style, frequency,
        order, count, complement, and gain are electively specified.
        Alternatively, a forward coefficient array and state array may be explicitly
        defined to ignore constraints.

        Frequency corresponds to a -3 dB frequency response normalized relative
        to Nyquist.

        Style is in ( 'Blackman', 'Hamming', 'Hann', 'Kaiser' ).

        * | 'Blackman' filters demonstrate low resolution and spectral leakage
          | with improved rate of attenuation.

        * | 'Hamming' filters demonstrate minimal nearest side lobe magnitude
          | response.

        * | 'Hann' filters demonstrate high resolution and spectral leakage.

        * | 'Kaiser' filters demonstrate flexible resolution and spectral
          | leakage dependent upon a beta value of a Bessel function of the
          | first kind, with beta equal to 7.0.

        Order must even to ensure a Type I form linear phase solution.

        Count is a quantity of filters of a specified order concatenated to
        form an aggregate frequency response in cascade form.

        Complement effectively constructs a mirror image of a specified
        frequency response.

    **Example**
        
        ::
        
            from diamondback import FirFilter
            import numpy

            # Create an instance with constraints.

            obj = FirFilter( style = 'Kaiser', frequency = 0.1, order = 32, count = 1 )

            # Create an instance with coefficients.

            obj = FirFilter( b = obj.b )

            # Estimate frequency response, group delay, and roots.

            y, f = obj.response( length = 8192, count = 1 )
            y, f = obj.delay( length = 8192, count = 1 )
            p, z = obj.roots( count = 1 )

            # Filter an incident signal.

            x = numpy.random.rand( 128 ) * 2.0 - 1.0
            obj.reset( x[ 0 ] )
            y = obj.filter( x )

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-01-23.
"""

from typing import List, Tuple, Union
import math
import numpy
import scipy.signal
import warnings

class FirFilter( object ) :

    """ Finite Impulse Response ( FIR ) filter.
    """

    __style = ( 'Blackman', 'Hamming', 'Hann', 'Kaiser' )

    @property
    def b( self ) :

        """ b : Union[ List, numpy.ndarray ] - forward coefficient.
        """

        return self._b

    @b.setter
    def b( self, b : Union[ List, numpy.ndarray ] ) :

        self._b = b

    @property
    def s( self ) :

        """ s : Union[ List, numpy.ndarray ] - state.
        """

        return self._s

    @s.setter
    def s( self, s : Union[ List, numpy.ndarray ] ) :

        self._s = s

    def __init__( self, style : str = '', frequency : float = 0.0, order : int = 0, count : int = 1, complement : bool = False, gain : float = 1.0,
                  b : Union[ List, numpy.ndarray ] = [ ], s : Union[ List, numpy.ndarray ] = [ ] ) -> None :

        """ Initialize.

            Specify constraints including style, frequency, and order.
            Alternatively, a forward coefficient array may be explicitly defined
            to ignore constraints.

            Labels should be used to avoid ambiguity between constraints and
            coefficients.

            Arguments :
                style : str - in ( 'Blackman', 'Hamming', 'Hann', 'Kaiser' ).
                frequency : float - frequency normalized to Nyquist in ( 0.0, 1.0 ).
                order : int - order per instance.
                count : int - instances per cascade.
                complement : bool - complement response.
                gain : float - gain.
                b : Union[ List, numpy.ndarray ] - forward coefficient.
                s : Union[ List, numpy.ndarray ] - state.
        """

        if ( not len( b ) ) :
            if ( ( not style ) or ( style not in FirFilter.__style ) ) :
                raise ValueError( f'style = {style} Expected Style in {FirFilter.__style}' )
            if ( ( frequency <= 0.0 ) or ( frequency >= 1.0 ) ) :
                raise ValueError( f'Frequency = {frequency} Expected Frequency in ( 0.0, 1.0 )' )
            if ( order < 0 ) :
                raise ValueError( f'Order = {order} Expected Order in [ 0, inf )' )
            if ( count < 1 ) :
                raise ValueError( f'Count = {count} Expected Count in [ 1, inf )' )
            if ( complement ) :
                frequency = 1.0 - frequency
            if ( style == 'Kaiser' ) :
                window = ( style.lower( ), 7.0 )
            else :
                window = style.lower( )
            beta, eps, error = 10.0, numpy.finfo( float ).eps, float( 'inf' )
            index, mu, zeta = 500 * ( 1 + ( count > 2 ) ), 2.5e-2, 1.0
            for _ in range( 0, index ) :
                with warnings.catch_warnings( ) :
                    warnings.simplefilter( 'ignore' )
                    v = scipy.signal.firwin( order + 1, zeta * frequency, None, window, True, True, 1.0 )
                    if ( numpy.isnan( v ).any( ) ) :
                        raise ValueError( f'V = {v}' )
                    x = numpy.exp( 1j * math.pi * frequency )
                    e = ( 2.0 ** ( -0.5 ) ) - ( abs( numpy.polyval( v, x ) ) ** count )
                    if ( abs( e ) < error ) :
                        b, error = v, abs( e )
                        if ( error < ( 10.0 * eps ) ) :
                            break
                    zeta = numpy.maximum( zeta + mu * math.tanh( beta * e ), eps )
            if ( complement ) :
                b *= numpy.array( [ ( ( -1.0 ) ** x ) for x in range( 0, len( b ) ) ] )
                b /= sum( b * numpy.array( [ ( ( -1.0 ) ** x ) for x in range( 0, len( b ) ) ] ) )
            b *= gain
        if ( ( not numpy.isscalar( b ) ) and ( not isinstance( b, numpy.ndarray ) ) ) :
            b = numpy.array( list( b ) )
        if ( not len( b ) ) :
            raise ValueError( f'B = {b}' )
        if ( ( not numpy.isscalar( s ) ) and ( not isinstance( s, numpy.ndarray ) ) ) :
            s = numpy.array( list( s ) )
        if ( len( b ) < len( s ) ) :
            b = numpy.concatenate( ( b, numpy.zeros( len( s ) - len( b ) ) ) )
        if ( len( s ) < len( b ) ) :
            s = numpy.concatenate( ( s, numpy.zeros( len( b ) - len( s ) ) ) )
        super( ).__init__( )
        self._b, self._s = numpy.array( b ), numpy.array( s, type( b[ 0 ] ) )

    def delay( self, length : int = 8192, count : int = 1 ) -> Tuple[ numpy.ndarray, numpy.ndarray ] :

        """ Estimates group delay and produces a reference signal.

            Arguments :
                length : int.
                count : int.

            Returns :
                y : numpy.ndarray - reference signal.
                f : numpy.ndarray - frequency normalized to Nyquist in [ -1.0, 1.0 ).
        """

        if ( length < 1 ) :
            raise ValueError( f'Length = {length} Expected Length in [ 1, inf )' )
        if ( count < 1 ) :
            raise ValueError( f'Count = {count} Expected Count in [ 1, inf )' )
        with warnings.catch_warnings( ) :
            warnings.simplefilter( 'ignore' )
            y, f = scipy.signal.group_delay( ( self.b, [ 1.0 ] ), length, True )[ 1 ], numpy.linspace( -1.0, 1.0 - 2.0 / length, length )
            y = numpy.concatenate( ( y[ len( y ) // 2 : ], y[ : len( y ) // 2 ] ) ) * count
        if ( length > 2 ) :
            y[ 0 ] = y[ 1 ] * 2.0 - y[ 2 ]
        return y, f

    def filter( self, x : Union[ List, numpy.ndarray ] ) -> numpy.ndarray :

        """ Filters an incident signal and produces a reference signal.

            Arguments :
                x : Union[ List, numpy.ndarray ] - incident signal.

            Returns :
                y : numpy.ndarray - reference signal.
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :
            x = numpy.array( list( x ) )
        if ( not len( x ) ) :
            raise ValueError( f'X = {x}' )
        y = numpy.zeros( len( x ), type( self.b[ 0 ] ) )
        for ii in range( 0, len( x ) ) :
            self.s[ 0 ] = x[ ii ]
            y[ ii ] = self.b.dot( self.s )
            if ( len( self.s ) > 1 ) :
                self.s[ 1 : ] = self.s[ : -1 ]
        return y

    def reset( self, x : Union[ complex, float ] ) -> None :

        """ Modifies a state to minimize edge effects by assuming persistent
            operation at a specified incident signal condition.

            Arguments :
                x : Union[ complex, float ] - incident signal.
        """

        if ( not numpy.isscalar( x ) ) :
            raise ValueError( f'X = {x}' )
        self.s.fill( x )

    def response( self, length = 8192, count = 1 ) -> Tuple[ numpy.ndarray, numpy.ndarray ] :

        """ Estimates frequency response and produces a reference signal.

            Arguments :
                length : int.
                count : int.

            Returns :
                y : numpy.ndarray - reference signal.
                f : numpy.ndarray - frequency normalized to Nyquist in [ -1.0, 1.0 ).
        """

        if ( length < 1 ) :
            raise ValueError( f'Length = {length} Expected Length in [ 1, inf )' )
        if ( count < 1 ) :
            raise ValueError( f'Count = {count} Expected Count in [ 1, inf )' )
        y, f = scipy.signal.freqz( self.b, [ 1.0, 0.0 ], length, True )[ 1 ], numpy.linspace( -1.0, 1.0 - 2.0 / length, length )
        y = numpy.concatenate( ( y[ len( y ) // 2 : ], y[ : len( y ) // 2 ] ) ) ** count
        return y, f

    def roots( self, count = 1 ) -> Tuple[ numpy.ndarray, numpy.ndarray ] :

        """ Estimates roots of a frequency response in poles and zeros.

            Arguments :
                count : int.
            Returns :
                p : numpy.ndarray - poles.
                z : numpy.ndarray - zeros.
        """

        z = numpy.tile( numpy.roots( self.b ), count )
        return numpy.zeros( count * ( len( self.b ) - 1 ) ), z[ numpy.argsort( abs( z ) ) ]



""" **Description**
        A derivative filter realizes a discrete difference equation which
        approximates a discrete derivative as a function of a forward coefficient
        array and a state array of a specified order, consuming an incident signal
        and producing a reference signal.  A derivative is approximated relative
        to a sample.  A derivative is electively approximated relative to a second
        by multiplying a reference signal by an absolute sampling frequency raised
        to a derivative power.

        .. math::
            y_{n} = \sum_{i = 0}^{N} b_{i} x_{n-i} = \sum_{i = 1}^{N} b_{i} s_{i,n} + b_{0} x_{n}
            s_{1,n+1} = x_{n}\qquad\quad s_{i,n+1} = s_{i-1,n}

        A frequency response is expressed as a function of a forward coefficient
        array.

        .. math::
            H_{z} = \sum_{i = 0}^{N} b_{i} z^{-i}

        A forward coefficient array and a state array of a specified order are
        defined, to satisfy specified constraints.  A derivative, and order are
        specified.

        .. math::
            y_{n} = f^{D}\ \\frac{x_{n}}{D}\quad\quad\quad\quad\scriptsize{ f = 1.0 }

        .. math::
            b_{1,1} = \scriptsize{ [ \matrix{ 1 & -1 } ] }

        .. math::
            b_{1,2} = \scriptsize{ [ \matrix{ 1 & 0 & -1 } ]\ \\frac{1}{2} }

        .. math::
            b_{1,4} = \scriptsize{ [ \matrix{ -1 & 8 & 0 & -8 & 1 } ]\ \\frac{1}{12} }

        .. math::
            b_{2,2} = \scriptsize{ [ \matrix{ 1 & -2 & 1 } ] }

        .. math::
            b_{2,4} = \scriptsize{ [ \matrix{ 1 & 0 & -2 & 0 & 1 } ]\ \\frac{1}{4} }

        .. math::
            b_{2,6} = \scriptsize{ [ \matrix{ -1 & 8 & 1 & -16 & 1 & 8 & -1 } ]\ \\frac{1}{24} }

        .. math::
            b_{2,8} = \scriptsize{ [ \matrix{ 1 & -16 & 64 & 16 & -130 & 16 & 64 & -16 & 1 } ]\ \\frac{1}{144} }

        .. math::
            b_{3,4} = \scriptsize{ [ \matrix{ 1 & -2 & 0 & 2 & -1 } ]\ \\frac{1}{2} }

        .. math::
            b_{3,6} = \scriptsize{ [ \matrix{ 1 & 0 & -3 & 0 & 3 & 0 & 1 } ]\ \\frac{1}{8} }

        .. math::
            b_{3,8} = \scriptsize{ [ \matrix{ -1 & 8 & 2 & -24 & 0 & 24 & -2 & -8 & 1 } ]\ \\frac{1}{48} }

    **Example**
        
        ::
        
            from diamondback import ComplexExponentialFilter, DerivativeFilter
            import numpy

            # Create an instance.

            obj = DerivativeFilter( derivative = 1, order = 2 )

            # Filter an incident signal.

            x = ComplexExponentialFilter( 0.0 ).filter( numpy.ones( 128 ) * 0.1 ).real
            obj.reset( x[ 0 ] )
            y = obj.filter( x )

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-02-06.
"""

from diamondback.filters.FirFilter import FirFilter
from typing import List, Union
import numpy

class DerivativeFilter( FirFilter ) :

    """ Derivative filter.
    """

    __b = { 1 : { 1 : numpy.array( [ 1.0, -1.0 ] ),
                  2 : numpy.array( [ 1.0, 0.0, -1.0 ] ) * ( 1.0 / 2.0 ),
                  4 : numpy.array( [ -1.0, 8.0, 0.0, -8.0, 1.0 ] ) * ( 1.0 / 12.0 ) },
            2 : { 2 : numpy.array( [ 1.0, -2.0, 1.0 ] ),
                  4 : numpy.array( [ 1.0, 0.0, -2.0, 0.0, 1.0 ] ) * ( 1.0 / 4.0 ),
                  6 : numpy.array( [ -1.0, 8.0, 1.0, -16.0, 1.0, 8.0, -1.0 ] ) * ( 1.0 / 24.0 ),
                  8 : numpy.array( [ 1.0, -16.0, 64.0, 16.0, -130.0, 16.0, 64.0, -16.0, 1.0 ] ) * ( 1.0 / 144.0 ) },
            3 : { 4 : numpy.array( [ 1.0, -2.0, 0.0, 2.0, -1.0 ] ) * ( 1.0 / 2.0 ),
                  6 : numpy.array( [ 1.0, 0.0, -3.0, 0.0, 3.0, 0.0, -1.0 ] ) * ( 1.0 / 8.0 ),
                  8 : numpy.array( [ -1.0, 8.0, 2.0, -24.0, 0.0, 24.0, -2.0, -8.0, 1.0 ] ) * ( 1.0 / 48.0 ) } }

    def __init__( self, derivative : int, order : int ) -> None :

        """ Initialize.

            Arguments :
                derivative : int - in [ 1, 3 ].
                order : int.
        """

        if ( derivative not in DerivativeFilter.__b ) :
            raise ValueError( f'Derivative = {derivative} Expected Derivative in {tuple( DerivativeFilter.__b.keys( ) )}' )
        if ( order not in DerivativeFilter.__b[ derivative ] ) :
            raise ValueError( f'Order = {order} Expected Order in {tuple( DerivativeFilter.__b[ derivative ].keys( ) )}' )
        super( ).__init__( b = DerivativeFilter.__b[ derivative ][ order ] )

    def filter( self, x : Union[ List, numpy.ndarray ] ) -> numpy.ndarray :

        """ Filters an incident signal and produces a reference signal.

            Arguments :
                x : Union[ List, numpy.ndarray ] - incident signal.

            Returns :
                y : numpy.ndarray - reference signal.
        """

        return super( ).filter( x )



""" **Description**
        A diversity model realizes the selection and retention of a state as a
        finite collection of observations extracted from an incident signal, to
        maximize a minimum distance between any members of a state, according to
        a specified style or distance metric.

        .. math::
            d_{k} = \min(\ d_{u,v}\ )\quad\quad u, v \\in [\ 0,\ M\ ),\ u \\neq v

        .. math::
            d_{k} \geq d_{n}\qquad \longrightarrow\qquad d_{n} = d_{k}

        A diversity model is an opportunistic unsupervised learning model which
        typically improves condition and numerical accuracy and reduces storage
        relative to alternative approaches including generalized linear inverse.

        A state array of a specified order is defined.  A stationary dimension is
        inferred.  A style and order are specified.

        Style is in ( 'Chebyshev', 'Euclidean', 'Geometric', 'Manhattan' ).

        * | 'Chebyshev' distance is an L-infinity norm, a maximum absolute difference
          | in any dimension.

        .. math::
            d_{u,v} = \max(\ |\ \\vec{x_{u}} - \\vec{x_{v}}\ |\ )

        * | 'Euclidean' distance is an L-2 norm, a square root of a sum of squared
          | differences in each dimension.

        .. math::
            d_{u,v} = \matrix{\sum_{i=0}^{N}(\ |\ \\vec{x_{u,i}} - \\vec{x_{v,i}}\ )^2|}^{0.5}

        * | 'Geometric' distance is a ordered root of a product of absolute differences
          | in each dimension.

        .. math::
            d_{u,v} = \prod_{i=0}^{N}{(\ |\ \\vec{x_{u,i}} - \\vec{x_{v,i}}\ |\ )}^{\\frac{1}{N}}

        * | 'Manhattan' distance is an L-1 norm, a sum of absolute differences in each
          | dimension.

        .. math::
            d_{u,v} = \sum_{i=0}^{N}{\ (\ |\ \\vec{x_{u}} - \\vec{x_{v}}\ |\ )\ }

    **Example**
      
        ::
        
            from diamondback import DiversityModel

            # Create an instance.

            obj = DiversityModel( style = 'Euclidean', order = 4 )

            # Learn an incident signal and extract a state.

            x = numpy.random.rand( 32, 2 )
            y = obj.learn( x )
            s = obj.s

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-02-08.
"""

from typing import List, Union
import numpy

class DiversityModel( object ) :

    """ Diversity model.
    """

    __distance = dict( Chebyshev = lambda x, y : max( abs( x - y ) ),
                       Euclidean = lambda x, y : sum( ( x - y ) ** 2 ) ** 0.5,
                       Geometric = lambda x, y : numpy.prod( abs( x - y ) ) ** ( 1.0 / len( x ) ),
                       Manhattan = lambda x, y : sum( abs( x - y ) ) )

    @property
    def s( self ) :

        """ s : Union[ List, numpy.ndarray ] - state.
        """

        return self._s

    @s.setter
    def s( self, s : Union[ List, numpy.ndarray ] ) :

        self._s = s

    def __init__( self, style : str, order : int ) -> None :

        """ Initialize.

            Arguments :
                style : str - in ( 'Chebyshev', 'Euclidean', 'Geometric', 'Manhattan' ).
                order : int.
        """

        if ( ( not style ) or ( style not in DiversityModel.__distance ) ) :
            raise ValueError( f'style = {style} Expected Style in {tuple( DiversityModel.__distance.keys( ) )}' )
        if ( order < 1 ) :
            raise ValueError( f'Order = {order} Expected Order in [ 1, inf )' )
        super( ).__init__( )
        self._distance = DiversityModel.__distance[ style ]
        self._diversity = 0.0
        self._s = numpy.zeros( ( order + 1, 0 ) )

    def clear( self ) -> None :

        """ Clears an instance.
        """

        self._diversity = 0.0
        self.s = numpy.zeros( ( self.s.shape[ 1 ], 0 ) )

    def learn( self, x : Union[ List, numpy.ndarray ] ) -> numpy.ndarray :

        """ Learns an incident signal and produces a reference signal.

            Arguments :
                x : Union[ List, numpy.ndarray ] - incident signal.

            Returns :
                y : numpy.ndarray - diversity.
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :
            x = numpy.array( list( x ) )
        if ( ( len( x.shape ) != 2 ) or ( not all( x.shape ) ) ) :
            raise ValueError( f'X = {x}' )
        if ( not self.s.shape[ 1 ] ) :
            self.s = numpy.zeros( ( self.s.shape[ 0 ], x.shape[ 1 ] ) ) + numpy.finfo( float ).max
        if ( x.shape[ 1 ] != self.s.shape[ 1 ] ) :
            raise ValueError( f'X = {x.shape} S = {s.shape}' )
        cc = 0
        for ii in range( 0, self.s.shape[ 0 ] ) :
            if ( numpy.isclose( self.s[ ii, 0 ], numpy.finfo( float ).max ) ) :
                break
            cc += 1
        y = numpy.zeros( x.shape[ 0 ] )
        for ii in range( 0, x.shape[ 0 ] ) :
            if ( cc < self.s.shape[ 0 ] ) :
                self.s[ cc, : ] = x[ ii, : ]
                cc += 1
            else :
                v, jj = self._diversity, -1
                for kk in range( 0, cc ) :
                    u, s = float( 'inf' ), numpy.array( self.s )
                    s[ kk, : ] = x[ ii, : ]
                    for uu in range( 0, cc - 1 ) :
                        for vv in range( uu + 1, cc ) :
                            d = self._distance( s[ uu, : ], s[ vv, : ] )
                            if ( d < u ) :
                                u = d
                    if ( u > v ) :
                        v, jj = u, kk
                if ( v > self._diversity ) :
                    self._diversity, self.s[ jj, : ] = v, x[ ii, : ]
            y[ ii ] = self._diversity
        return y



""" **Description**
        A window filter realizes a discrete difference equation as a function
        of a forward coefficient array of a specified order, consuming an
        incident signal and producing a reference signal.

        .. math::
            y_{n} = b_{n}\ x_{n}

        A forward coefficient array of a specified order is defined.  A
        style, order, and normalization are specified.

        Style is in ( 'Blackman', 'Hamming', 'Hann', 'Kaiser' ).

        * | 'Blackman' filters demonstrate low resolution and spectral leakage
          | with improved rate of attenuation.

        * | 'Hamming' filters demonstrate minimal nearest side lobe magnitude
          | response.

        * | 'Hann' filters demonstrate high resolution and spectral leakage.

        * | 'Kaiser' filters demonstrate flexible resolution and spectral
          | leakage dependent upon a beta value of a Bessel function of the
          | first kind, with beta equal to 7.0.

        Normal condition scales a forward coefficient array to electively
        compensate for energy loss.

        .. math::
            b_{n} = b_{n}\ \\frac{ N }{ \sum_{0}^{N-1}\ |\ b_{n}\ |}

    **Example**
     
        ::
        
            from diamondback import ComplexExponentialFilter, WindowFilter
            import numpy

            # Create an instance.

            obj = WindowFilter( style = 'Hann', order = 15, normal = True )

            # Filter an incident signal.

            x = ComplexExponentialFilter( 0.0 ).filter( numpy.ones( len( obj.b ) ) * 0.1 ).real
            y = obj.filter( x )

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-04-13.
"""

from typing import List, Union
import numpy
import scipy.signal

class WindowFilter( object ) :

    """ Window filter.
    """

    __style = ( 'Blackman', 'Hamming', 'Hann', 'Kaiser' )

    @property
    def b( self ) :

        """ b : Union[ List, numpy.ndarray ] - forward coefficient.
        """

        return self._b

    def __init__( self, style : str, order : int, normal : bool = True  ) -> None :

        """ Initialize.

            Arguments :
                style : str - in ( 'Blackman', 'Hamming', 'Hann', 'Kaiser' ).
                order : int.
                normal : bool.
        """

        if ( ( not style ) or ( style not in WindowFilter.__style ) ) :
            raise ValueError( f'style = {style} Expected Style in {WindowFilter.__style}' )
        if ( order < 0 ) :
            raise ValueError( f'Order = {order} Expected Order in [ 0, inf )' )
        if ( style == 'Kaiser' ) :
            window = ( style.lower( ), 7.0 )
        else :
            window = style.lower( )
        super( ).__init__( )
        b = scipy.signal.get_window( window, order + 1, False )
        if ( normal ) :
            b *= ( order + 1 ) / sum( abs( b ) )
        self._b = numpy.array( b )

    def filter( self, x : Union[ List, numpy.ndarray ] ) -> numpy.ndarray :

        """ Filters an incident signal and produces a reference signal.

            Arguments :
                x : Union[ List, numpy.ndarray ] - incident signal.

            Returns :
                y : numpy.ndarray - reference signal.
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :
            x = numpy.array( list( x ) )
        if ( ( len( x.shape ) != 1 ) or ( len( x ) != len( self.b ) ) ) :
            raise ValueError( f'X = {x}' )
        return self.b * x



""" **Description**
        A complex transform converts a three-phase real signal to a complex
        signal, or converts a complex signal to a three-phase real signal, in
        equivalent and reversible representations.  A neutral condition is
        specified.

        Singleton.

        .. math::
            \gamma = \scriptsize{ \matrix{ 1 & \\frac{-1}{2} & \\frac{-1}{2}\cr 0 & \\frac{3^{0.5}}{2} & \\frac{-3^{0.5}}{2}\cr 1 & 1 & 1 } }

        .. math::
            \\alpha = \matrix{\\frac{1}{3}}^{0.5}\ e^{\ -j\ \\frac{\pi}{6}\ }

        Complex - Three-Phase.

        .. math::
            y_{n} = \matrix{ x^{T}_{n}\ \gamma\ }\ \scriptsize{[\ \matrix{ 1 & j & 0 }\ ]^{T}}

        .. math::
            \overline{\scriptsize{Neutral}}\ \qquad\longrightarrow\qquad y_{n} = y_{n}\ \\alpha

        Three-Phase - Complex.

        .. math::
            \overline{\scriptsize{Neutral}}\ \qquad\longrightarrow\qquad x_{n} = \\frac{x_{n}}{\\alpha}

        .. math::
            y_{n} = \matrix{\ \matrix{ \gamma^{T}\gamma }^{-1}\ \gamma^{T}\ \scriptsize{[\ \matrix{ real(\ x_{n}\ ) & imag(\ x_{n}\ ) & 0 }\ ]^{T}}}^{T}

    **Example**
       
        ::
        
            from diamondback import ComplexExponentialFilter, ComplexTransform
            import numpy

            x = ComplexExponentialFilter( 0.0 ).filter( numpy.linspace( -1.0e-4, 1.0e-4, 128 ) + 0.1 )

            # Transform an incident signal, forward and inverse.

            y = ComplexTransform.transform( x, neutral = True )
            z = ComplexTransform.transform( y, neutral = True )

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-01-26.
"""

from typing import List, Union
import math
import numpy

class ComplexTransform( object ) :

    """ Complex transform.
    """

    __coefficient = ( 2.0 / 3.0 ) * numpy.array( [ [ 1.0, -0.5, -0.5 ],
                                                   [ 0.0, 0.5 * ( 3.0 ** 0.5 ), -0.5 * ( 3.0 ** 0.5 ) ],
                                                   [ 1.0, 1.0, 1.0 ] ] )

    __gain = ( ( 1.0 / 3.0 ) ** 0.5 ) * numpy.exp( -1j * math.pi / 6.0 )

    @classmethod
    def transform( cls, x : Union[ List, numpy.ndarray ], neutral : bool = True ) -> numpy.ndarray :

        """ Transforms a real three-phase or complex incident signal into a complex
            or three-phase reference signal.

            Arguments :
                x : Union[ List, numpy.ndarray ] - incident signal.
                neutral : bool.

            Returns :
                y : numpy.ndarray - reference signal.
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :
            x = numpy.array( list( x ) )
        if ( ( len( x.shape ) > 2 ) or ( len( x ) == 0 ) ) :
            raise ValueError( f'X = {x}' )
        if ( len( x.shape ) < 2 ) :
            rows, cols = 1, x.shape[ 0 ]
        else :
            rows, cols = x.shape
        if ( ( rows not in ( 1, 3 ) ) or ( cols < 1 ) ) :
            raise ValueError( f'Rows = {rows} Columns = {cols} Expected Rows in ( 1, 3 ) and Columns in [ 1, inf )' )
        if ( rows == 1 ) :
            v = x
            if ( not neutral ) :
                v = x / ComplexTransform.__gain
            y = numpy.matmul( numpy.linalg.inv( ComplexTransform.__coefficient ), numpy.array( [ v.real, v.imag, numpy.zeros( cols ) ] ) )
        else :
            v = numpy.matmul( ComplexTransform.__coefficient, x )
            y = v[ 0, : ] + 1j * v[ 1, : ]
            if ( not neutral ) :
                y *= ComplexTransform.__gain
        return y



""" **Description**
        A Proportional Integral Derivative ( PID ) filter realizes a discrete
        difference equation as a function of a forward coefficient array and a
        state array of a static order.  A forward coefficient array applies a
        gain to proportional, integral, and derivative representations of an
        incident signal, producing a reference signal.  An integral limit is
        specified, preventing integral saturation which may adversely affect
        control stability and latency.

        .. math::
            y_{n} = b_{0}\ x_{n} + b_{1}\max(\ min( \sum_{0}^{n}\ x_{n},\ limit\ ),\ -limit\ ) + b_{2}\ \\frac{d}{dn}(\ x_{n}\ )

    **Example**
        
        ::
        
            from diamondback import ComplexExponentialFilter, PidFilter
            import numpy

            # Create an instance.

            obj = PidFilter( b = numpy.array( [ 0.1, 5.0e-2, 0.0 ] ) )

            # Filter an incident signal.

            x = ComplexExponentialFilter( 0.0 ).filter( numpy.linspace( -1.0e-4, 1.0e-4, 128 ) * 0.1 ).real
            y = obj.filter( x )

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-01-31.
"""

from diamondback.filters.FirFilter import FirFilter
from typing import List, Union
import numpy

class PidFilter( FirFilter ) :

    """ Proportional Integral Derivative ( PID ) filter.
    """

    @property
    def limit( self ) :

        """ limit : float.
        """

        return self._limit

    @limit.setter
    def limit( self, limit : float ) :

        if ( limit < 0.0 ) :
            raise ValueError( f'Limit = {limit} Expected Limit in ( 0.0, inf )' )
        self._limit = limit

    def __init__( self, b : Union[ List, numpy.ndarray ] ) -> None :

        """ Initialize.

            Arguments :
                b : Union[ List, numpy.ndarray ] - forward coefficient.
        """

        if ( ( not numpy.isscalar( b ) ) and ( not isinstance( b, numpy.ndarray ) ) ) :
            b = numpy.array( list( b ) )
        if ( len( b ) != 3 ) :
            raise ValueError( f'B = {b}' )
        super( ).__init__( b = b, s = numpy.zeros( len( b ) ) )
        self._limit = float( 'inf' )

    def filter( self, x : Union[ List, numpy.ndarray ] ) -> numpy.ndarray :

        """ Filters an incident signal and produces a reference signal.

            Arguments :
                x : Union[ List, numpy.ndarray ] - incident signal.

            Returns :
                y : numpy.ndarray - reference signal.
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :
            x = numpy.array( list( x ) )
        if ( not len( x ) ) :
            raise ValueError( f'X = {x}' )
        y = numpy.zeros( len( x ), type( self.b[ 0 ] ) )
        for ii in range( 0, len( x ) ) :
            self.s[ 2 ] = x[ ii ] - self.s[ 0 ]
            if ( abs( self.s[ 1 ] + x[ ii ] ) < self.limit ) :
                self.s[ 1 ] += x[ ii ]
            self.s[ 0 ] = x[ ii ]
            y[ ii ] = self.b.dot( self.s )
        return y



""" **Description**
        A complex band pass filter produces a complex exponential incident
        signal at a specified normalized frequency and adapts a forward complex
        coefficient to produce a reference signal, which estimates a component
        of interest in a primary signal.  A normalized frequency and rate of
        adaptation are specified.

        .. math::
            x_{n} = e^{\ j\ \pi\ \phi_{n}}

        .. math::
            \phi_{n+1} = \phi_{n} + f_{n}

        .. math::
            y_{n} = b_{n} x_{n}

        .. math::
            e_{n} = d_{n} - y_{n}

        .. math::
            b_{n+1} = b_{n} + \mu e_{n} x_{n}^{*}

    **Example**
        
        ::
        
            from diamondback import ComplexBandPassFilter, ComplexExponentialFilter
            import numpy

            frequency = 0.1
            x = numpy.linspace( -1.0e-4, 1.0e-4, 128 ) + frequency

            # Create a primary signal.

            d = ComplexExponentialFilter( phase = numpy.random.rand( 1 )[ 0 ] * 2.0 - 1.0 ).filter( x )

            # Create an instance.

            obj = ComplexBandPassFilter( frequency = frequency, rate = 5.0e-2 )

            # Filter a primary signal.

            obj.reset( d[ 0 ] )
            y, e, b = obj.filter( d )

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-01-31.
"""

from diamondback.filters.ComplexExponentialFilter import ComplexExponentialFilter
from diamondback.filters.FirFilter import FirFilter
from typing import List, Tuple, Union
import numpy
import scipy

class ComplexBandPassFilter( FirFilter ) :

    """ Complex band pass filter.
    """

    @property
    def frequency( self ) :

        """ frequency : float - frequency normalized to Nyquist in [ -1.0, 1.0 ].
        """

        return self._frequency

    @frequency.setter
    def frequency( self, frequency : float ) :

        if ( ( frequency < -1.0 ) or ( frequency > 1.0 ) ) :
            raise ValueError( f'Frequency = {frequency} Expected Frequency in [ -1.0, 1.0 ]' )
        self._frequency = frequency

    @property
    def rate( self ) :

        """ rate : float - in [ 0.0, 1.0 ].
        """

        return self._rate

    @rate.setter
    def rate( self, rate : float ) :

        if ( ( rate < 0.0 ) or ( rate > 1.0 ) ) :
            raise ValueError( f'Rate = {rate} Expected Rate in [ 0.0, 1.0 ]' )
        self._rate = rate

    def __init__( self, frequency : float, rate : float ) -> None :

        """ Initialize.

            Arguments :
                frequency : float - frequency normalized to Nyquist in [ -1.0, 1.0 ).
                rate : float - in [ 0.0, 1.0 ].
        """

        if ( ( frequency < -1.0 ) or ( frequency > 1.0 ) ) :
            raise ValueError( f'Frequency = {frequency} Expected Frequency in [ -1.0, 1.0 ]' )
        if ( ( rate < 0.0 ) or ( rate > 1.0 ) ) :
            raise ValueError( f'Rate = {rate} Expected Rate in [ 0.0, 1.0 ]' )
        super( ).__init__( b = numpy.array( [ numpy.finfo( float ).eps + 0j ] ), s = numpy.zeros( 1, complex ) )
        self._complexexponentialfilter = ComplexExponentialFilter( )
        self._frequency, self._rate = frequency, rate

    def filter( self, d : Union[ List, numpy.ndarray ] ) -> Tuple[ numpy.ndarray, numpy.ndarray, numpy.ndarray ] :

        """ Filters a primary signal and produces a reference signal.

            Signals are Hilbert transformed to complex as necessary.

            Arguments :
                d : Union[ List, numpy.ndarray ] - primary signal.

            Returns :
                y : numpy.ndarray - reference signal.
                e : numpy.ndarray - error signal.
                b : numpy.ndarray - forward coefficient.
        """

        if ( ( not numpy.isscalar( d ) ) and ( not isinstance( d, numpy.ndarray ) ) ) :
            d = numpy.array( list( d ) )
        if ( not len( d ) ) :
            raise ValueError( f'D = {d}' )
        if ( not numpy.iscomplex( d ).any( ) ) :
            d = scipy.signal.hilbert( d )
        x = self._complexexponentialfilter.filter( numpy.ones( len( d ) ) * self.frequency )
        y, e, b = numpy.zeros( len( x ), complex ), numpy.zeros( len( x ), complex ), numpy.zeros( len( x ), complex )
        for ii in range( 0, len( x ) ) :
            y[ ii ] = x[ ii ] * self.b[ 0 ]
            e[ ii ] = d[ ii ] - y[ ii ]
            b[ ii ] = self.b[ 0 ]
            self.b[ 0 ] += self.rate * e[ ii ] * numpy.conjugate( x[ ii ] )
        return y, e, b



""" **Description**
        Test unit models.

    **Example**
     
        ::
        
            pytest --capture=no --verbose
    
    **License**
        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.
    
    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-04-03.
"""

from diamondback import ComplexExponentialFilter, WindowFilter
from diamondback import ComplexTransform, FourierTransform, PowerSpectrumTransform, WaveletTransform
from diamondback import ZTransform
import math
import numpy

class Test( object ) :

    """ Test.
    """

    def test_ComplexTransform( self ) :

        """ Test ComplexTransform.
        """

        count = 32
        frequency = 0.1
        x = ComplexExponentialFilter( 0.5 ).filter( numpy.linspace( -1.0e-4, 1.0e-4, count ) + frequency )
        for ii in range( 0, 2 ) :
            y = ComplexTransform.transform( x, not ii )
            assert y.shape == ( 3, count )
            assert not isinstance( y[ 0 ], complex )
            z = ComplexTransform.transform( y, not ii )
            assert z.shape == ( count, )
            assert isinstance( z[ 0 ], complex )
            assert numpy.allclose( z, x )

    def test_FourierTransform( self ) :

        """ Test FourierTransform.
        """

        count = 128
        frequency = ( 0.12, 0.23 )
        b = numpy.ones( count )
        for ii in range( 0, 4 ) :
            x = numpy.random.rand( ) * ComplexExponentialFilter( numpy.random.rand( ) ).filter( numpy.linspace( frequency[ 0 ], frequency[ 1 ], count ) )
            if ( ii % 2 ) :
                x = x.real
            if ( ii >= 2 ) :
                b = WindowFilter( 'Hann', count - 1 ).b
            y, f = FourierTransform.transform( x, b )
            assert ( ( len( y ) == count ) and ( len( f ) == count ) )
            assert isinstance( y[ 0 ], complex )
            assert numpy.isclose( f[ 0 ], -1.0 ) and numpy.isclose( f[ -1 ], 1.0 - 2.0 / count )
            h = numpy.argmax( abs( y[ count // 2 : ] ) )
            assert ( h >= numpy.argmin( abs( f[ count // 2 : ] - frequency[ 0 ] ) ) ) and ( h <= numpy.argmin( abs( f[ count // 2 : ] - frequency[ 1 ] ) ) )
            z = FourierTransform.transform( y, b, True )[ 0 ]
            if ( ii % 2 ) :
                z = z.real
            assert numpy.allclose( z[ 1 : -1 ], x[ 1 : -1 ] )

    def test_PowerSpectrumTransform( self ) :

        """ Test PowerSpectrumTransform.
        """

        count = 128
        frequency, index = ( 0.12, 0.23 ), 64
        b = WindowFilter( 'Hann', count - 1 ).b
        for ii in range( 0, 2 ) :
            x = numpy.random.rand( ) * ComplexExponentialFilter( numpy.random.rand( ) ).filter( numpy.linspace( frequency[ 0 ], frequency[ 1 ], count * 8 ) )
            if ( ii > 0 ) :
                x = x.real
            y, f = PowerSpectrumTransform.transform( x, b, index, False )
            assert ( ( len( y ) == ( count // 2 ) ) and ( len( f ) == ( count // 2 ) ) )
            assert not isinstance( y[ 0 ], complex )
            assert numpy.isclose( f[ 0 ], 0.0 ) and numpy.isclose( f[ -1 ], 1.0 - 2.0 / count )

    def test_WaveletTransform( self ) :

        """ Test WaveletTransform.
        """

        count = 3
        h = numpy.array( [  0.235233603892082,  0.570558457915722, 0.325182500263116, -0.095467207784164,
                           -0.060416104155198, 0.024908749868442 ] )
        g = numpy.array( [ -0.024908749868442, -0.060416104155198, 0.095467207784164,  0.325182500263116,
                           -0.570558457915722, 0.235233603892082 ] )
        obj = WaveletTransform( 'Daubechies', len( h ) - 1 )
        assert numpy.allclose( obj.b[ 0 ][ 0 ], h )
        assert numpy.allclose( obj.b[ 0 ][ 1 ], g )
        h = numpy.array( [ 0.024908749868442, -0.060416104155198, -0.095467207784164, 0.325182500263116,
                           0.570558457915722,  0.235233603892082 ] )
        g = numpy.array( [ 0.235233603892082, -0.570558457915722,  0.325182500263116, 0.095467207784164,
                           -0.060416104155198, -0.024908749868442 ] )
        assert numpy.allclose( obj.b[ 1 ][ 0 ], h )
        assert numpy.allclose( obj.b[ 1 ][ 1 ], g )
        h = numpy.array( [ 0.5, 0.5 ] )
        g = numpy.array( [ 0.5, -0.5 ] )
        obj = WaveletTransform( 'Haar', len( h ) - 1 )
        assert numpy.allclose( obj.b[ 0 ][ 0 ], h )
        assert numpy.allclose( obj.b[ 0 ][ 1 ], g )
        h = numpy.array( [ 0.5, 0.5 ] )
        g = numpy.array( [ -0.5, 0.5 ] )
        assert numpy.allclose( obj.b[ 1 ][ 0 ], h )
        assert numpy.allclose( obj.b[ 1 ][ 1 ], g )
        u = numpy.array( [ -0.677118658500953,  0.345009188833763,  0.118882064536892, -0.537364409907213,
                           -0.250663848240361,  0.491955501906881, -0.545336874797487,  0.390418096834100,
                           -0.271946313073118,  0.047745751406263,  0.301454810260592,  0.138563567406933,
                           -0.215817816000671, -0.271946313073118,  0.047745751406264,  0.301454810260592,
                           -0.154508497187474, -0.110615871041237, -0.024471741852423,  0.071019760960103,
                            0.139384128958763,  0.154508497187474,  0.110615871041237,  0.024471741852423,
                           -0.071019760960103, -0.139384128958763, -0.154508497187474, -0.110615871041237,
                           -0.024471741852423,  0.071019760960104,  0.139384128958763,  0.154508497187474 ] )
        x = ComplexExponentialFilter( 0.5 ).filter( numpy.ones( len( u ) ) * 0.1 ).real
        y = obj.transform( x, count, False )
        assert numpy.allclose( y, u )
        z = obj.transform( y, count, True )
        assert numpy.allclose( z, x )
        x = numpy.zeros( ( len( u ), len( u ) ) )
        for ii in range( 0, x.shape[ 0 ] ) :
            x[ ii, : ] = numpy.random.rand( ) * ComplexExponentialFilter( numpy.random.rand( ) * 2.0 - 1.0 ).filter( numpy.ones( len( u ) ) * numpy.random.rand( ) * 0.5 ).real
        z = obj.transform( obj.transform( x, count, False ), count, True )
        assert numpy.allclose( z, x )

    def test_ZTransform( self ) :

        """ Test ZTransform.
        """

        frequency, ripple = 0.05, 0.125
        u = numpy.array( [  0.000000000000000, 1.868522837482727, -1.863713887857194, 0.948381518868702,
                           -0.212543935987133 ] )
        v = numpy.array( [ 0.016209591718306, 0.064838366873225, 0.097257550309837, 0.064838366873225,
                           0.01620959171830 ] )
        s = numpy.array( [ numpy.exp( 1j * math.pi * x / ( ( len( u ) - 1 ) * 2 ) ) for x in range( 1, ( len( u ) - 1 ) * 2, 2 ) ] )
        t = math.asinh( 1.0 / ( ( 10.0 ** ( 0.1 * ripple ) - 1.0 ) ** 0.5 ) ) / ( len( u ) - 1 )
        a = ( numpy.poly( ( -math.sinh( t ) * s.imag + 1j * math.cosh( t ) * s.real ) * 2.0 * math.pi ) ).real
        a /= a[ -1 ]
        b = [ 1.0 ]
        a, b = ZTransform.transform( a, b, frequency, True )
        b = numpy.poly( -numpy.ones( ( len( u ) - 1 ) ) )
        b *= ( 1.0 - sum( a ) ) / sum( b )
        assert numpy.allclose( a, u )
        assert numpy.allclose( b, v )
        frequency = 0.025
        u = numpy.array( [ 0.000000000000000, 4.913032434807922, -10.502699472790392, 12.647757190438988,
                           -9.256183138855487, 4.112750158020544, -1.0263940516247210, 0.110901278364195 ] )
        v = numpy.array( [ 0.000000000160721, 0.002515288055351, 0.103543670353901, 0.397913857193261,
                           0.290734727994417, 0.040371147586715, 0.000522947605366, 0.000000000000000 ] ) * 1.0e-3
        s, a = numpy.ones( 1 ), numpy.ones( 2 )
        for ii in range( 2, len( u ) ) :
            x = numpy.concatenate( ( s, numpy.zeros( 2 ) ) ) + numpy.concatenate( ( [ 0.0 ], ( ( 2.0 * ii ) - 1.0 ) * a ) )
            s, a = a, x
        a /= a[ -1 ]
        b = [ 1.0 ]
        a, b = ZTransform.transform( a, b, frequency, False )
        assert numpy.allclose( a, u )
        assert numpy.allclose( b, v )



""" **Description**
        An integral filter realizes a discrete difference equation which
        approximates a discrete integral as a function of a recursive coefficient
        array, a forward coefficient array, and a state array of a specified order,
        consuming an incident signal and producing a reference signal.  An integral
        is approximated relative to a sample.  An integral is electively approximated
        relative to a second by dividing a reference signal by an absolute sampling
        frequency.

        .. math::
            y_{n} = \sum_{i = 1}^{N} a_{i} y_{n-i} + \sum_{i = 0}^{N} b_{i} x_{n-i} = \sum_{i = 1}^{N} (\ a_{i} b_{0} + b_{i}\ ) s_{i,n} + b_{0} x_{n}\qquad a_{0} = 0

        .. math::
            s_{1,n+1} = \sum_{i = 1}^{N} a_{i} s_{i,n} + x_{n}\qquad\quad s_{i,n+1} = s_{i-1,n}

        A frequency response is expressed as a function of a recursive coefficient
        array and a forward coefficient array.

        .. math::
            H_{z} = \\frac{\sum_{i = 0}^{N} b_{i} z^{-i}}{{1 - \sum_{i = 1}^{N} a_{i} z^{-i}}}

        A recursive coefficient array, forward coefficient array, and state array
        of a specified order are defined to satisfy specified constraints.  An
        instance and order are specified.

        .. math::
            y_{n} = \\frac{1}{f}\ \sum_{i=0}^{N} x_{n}\quad\quad\quad\quad\scriptsize{ f = 1.0 }

        .. math::
            \matrix{ a_{1,0} = \scriptsize{ [ \matrix{ 0 & 1 } ] } & b_{1,0} = \scriptsize{ [ \matrix{ 1 } ] } }\quad\quad\scriptsize{ Rectangular }

        .. math::
            \matrix{ a_{1,1} = \scriptsize{ [ \matrix{ 0 & 1 } ] } & b_{1,1} = \scriptsize{ [ \matrix{ 1 & 1 } ]\ \\frac{1}{2} } }\quad\quad\scriptsize{ Trapezoidal }

        .. math::
            \matrix{ a_{1,2} = \scriptsize{ [ \matrix{ 0 & 1 } ] } & b_{1,2} = \scriptsize{ [ \matrix{ 1 & 4 & 1 } ]\ \\frac{1}{6} } }\quad\quad\scriptsize{ Simpson\ 2 }

        .. math::
            \matrix{ a_{1,3} = \scriptsize{ [ \matrix{ 0 & 1 } ] } & b_{1,3} = \scriptsize{ [ \matrix{ 1 & 3 & 3 & 1 } ]\ \\frac{1}{8} } }\quad\quad\scriptsize{ Simpson\ 3 }

        .. math::
            \matrix{ a_{1,4} = \scriptsize{ [ \matrix{ 0 & 1 } ] } & b_{1,4} = \scriptsize{ [ \matrix{ 7 & 32 & 12 & 32 & 7 } ]\ \\frac{1}{90} } }\quad\quad\scriptsize{ Newton\ Coats }

    **Example**
       
        ::
        
            from diamondback import ComplexExponentialFilter, IntegralFilter
            import numpy

            # Create an instance.

            obj = IntegralFilter( order = 2 )

            # Filter an incident signal.

            x = ComplexExponentialFilter( 0.0 ).filter( numpy.ones( 128 ) * 0.1 ).real
            y = obj.filter( x )

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-02-06.
"""

from diamondback.filters.IirFilter import IirFilter
from typing import List, Union
import numpy

class IntegralFilter( IirFilter ) :

    """ Integral filter.
    """

    __b = ( numpy.array( [ 1.0 ] ),
            numpy.array( [ 1.0, 1.0 ] ) * ( 1.0 / 2.0 ),
            numpy.array( [ 1.0, 4.0, 1.0 ] ) * ( 1.0 / 6.0 ),
            numpy.array( [ 1.0, 3.0, 3.0, 1.0 ] ) * ( 1.0 / 8.0 ),
            numpy.array( [ 7.0, 32.0, 12.0, 32.0, 7.0 ] ) * ( 1.0 / 90.0 ) )

    def __init__( self, order : int ) -> None :

        """ Initialize.

            Arguments :
                order : int.
        """

        if ( ( order < 0 ) or ( order >= len( IntegralFilter.__b ) ) ) :
            raise ValueError( f'Order = {order} Expected Order in [ 0, {len( IntegralFilter.__b )} )' )
        super( ).__init__( a = numpy.array( [ 0.0, 1.0 ] ), b = IntegralFilter.__b[ order ] )

    def filter( self, x : Union[ List, numpy.ndarray ] ) -> numpy.ndarray :

        """ Filters an incident signal and produces a reference signal.

            Arguments :
                x : Union[ List, numpy.ndarray ] - incident signal.

            Returns :
                y : numpy.ndarray - reference signal.
        """

        return super( ).filter( x )



""" **Description**
        A Fourier transform converts a real or complex discrete-time incident
        signal to a complex discrete-frequency reference signal, or a complex
        discrete-frequency incident signal to a real or complex discrete-time
        reference signal, in equivalent and reversible representations.  A
        forward coefficient array is specified to define a window filter.

        Singleton.

        .. math::
            y_{k} = \\frac{1}{N}\ \sum_{n = 0}^{N-1} b_{n} x_{n} e^{\\frac{\ -j\ \pi\ k \ n}{N}} = \\frac{1}{N}\ fft(\ x_{n}\ )

        .. math::
            x_{n} = \\frac{N}{b_{n}}\ \sum_{k = 0}^{N-1} y_{k} e^{\\frac{\ j\ \pi\ k \ n}{N}} = \\frac{N}{b_{n}}\ ifft(\ y_{k}\ )

        A Fourier transform is normalized by incident signal length and forms
        a contiguous sequence corresponding to a linear and increasing
        normalized frequency.

        .. math::
            f_{k} = -1\ + \ 2\ \\frac{k}{N}

        An incident signal length is inversely proportional to a normalized
        frequency resolution.

            N = \\frac{2}{R}

    **Example**
    
        ::
        
            from diamondback import ComplexExponentialFilter, FourierTransform
            import numpy

            x = ComplexExponentialFilter( 0.0 ).filter( numpy.linspace( 0.12, 0.23, 128 ) ) * numpy.random.rand( 1 )[ 0 ]
            b = WindowFilter( 'Hann', len( x ) - 1 ).b

            # Transform an incident signal, forward and inverse.

            y, f = FourierTransform.transform( x, b = b, inverse = False )
            z = FourierTransform.transform( y, b = b, inverse = True )[ 0 ]

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-04-12.
"""

from typing import List, Tuple, Union
import numpy

class FourierTransform( object ) :

    """ Fourier transform.
    """

    @staticmethod
    def transform( x : Union[ List, numpy.ndarray ], b : Union[ List, numpy.ndarray ], inverse : bool = False ) -> Tuple[ numpy.ndarray, numpy.ndarray ] :

        """ Transforms a real or complex discrete-time incident signal to a
            complex discrete-frequency reference signal, or performs an
            inverse transform.  Indices definition depends upon an inverse
            condition.  Forward transform indices define normalized frequency.
            Inverse transform indices define an integral sequence.

            Arguments :
                x : Union[ List, numpy.ndarray ] - incident signal.
                b : Union[ List, numpy.ndarray ] - forward coefficient.
                inverse : bool.

            Returns :
                y : numpy.ndarray - reference signal.
                f : numpy.ndarray - frequency normalized to Nyquist in [ -1.0, 1.0 ).
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :
            x = numpy.array( list( x ) )
        if ( not len( x ) ) :
            raise ValueError( f'X = {x}' )
        if ( ( not numpy.isscalar( b ) ) and ( not isinstance( b, numpy.ndarray ) ) ) :
            b = numpy.array( list( b ) )
        if ( ( len( b ) > len( x ) ) or ( numpy.isclose( b, 0.0 ).all( ) ) ) :
            raise ValueError( f'B = {b}' )
        u = numpy.array( x[ : len( b ) ] )
        v = numpy.array( b )
        if ( inverse ) :
            if ( ( numpy.isclose( v[ 0 ], 0.0 ) ) or ( numpy.isclose( v[ -1 ], 0.0 ) ) ) :
                v[ 0 ], v[ -1 ] = v[ 1 ], v[ -2 ]
            y = numpy.concatenate( ( u[ len( u ) // 2 : ], u[ : len( u ) // 2 ] ) )
            y, f = numpy.fft.ifft( y ) * len( y ) / v, numpy.linspace( 0, len( y ) - 1, len( y ) )
        else :
            y, f = numpy.fft.fft( v * u ) / len( u ), numpy.linspace( -1.0, 1.0 - 2.0 / len( u ), len( u ) )
            y = numpy.concatenate( ( y[ len( y ) // 2 : ], y[ : len( y ) // 2 ] ) )
        return y, f



""" **Description**
        Initialize.

    **License**
        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-03-22.
"""

from .Log import Log
from .RestClient import RestClient
from .Serial import Serial

__all__ = [ 'Log', 'RestClient', 'Serial' ]



""" **Description**
        Diamondback setup.

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.
    
    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-01-23.
"""

from setuptools import setup

setup( )



""" **Description**
        Nox project management.

    **Example**
        
        ::
        
            nox --list
            nox --sessions build clean docs image login notebook push status tag tests
    
    **License**
        © 2019 - 2023 Schneider Electric Industries SAS. All rights reserved.
    
    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2020-10-12.
"""

import glob
import nox
import os
import pathlib
import requests
import shutil
import time

nox.options.sessions = [ 'build', 'docs', 'tests', 'push' ]

repository = pathlib.Path.cwd( ).name

@nox.session( venv_backend = 'none' )
def build( session ) -> None :

    """ Build distribution.
    """

    if ( pathlib.Path( 'setup.py' ).is_file( ) ) :
        shutil.rmtree( 'dist', ignore_errors = True )
        session.run( 'python', '-m', 'build', '-s', '-w' )
        shutil.rmtree( 'build', ignore_errors = True )
        session.run( 'git', 'add', str( pathlib.Path.cwd( ) / 'dist' / '*' ) )

@nox.session( venv_backend = 'none' )
def clean( session ) -> None :

    """ Clean repository.
    """

    for x in ( '.mypy_cache', '.nox', '.pytest_cache', 'build', 'dist', 'docs' ) :
        shutil.rmtree( x, ignore_errors = True )
    for x in [ x for x in glob.glob( f'**{str( pathlib.Path( "/" ) )}', recursive = True ) if ( '__pycache__' in x ) ] :
        shutil.rmtree( x, ignore_errors = True )

@nox.session( venv_backend = 'none' )
def docs( session ) -> None :

    """ Build documentation.
    """

    if ( pathlib.Path( 'templates' ).is_dir( ) ) :
        shutil.rmtree( 'docs', ignore_errors = True )
        ( pathlib.Path.cwd( ) / 'docs' ).mkdir( )
        session.run( 'sphinx-apidoc', '--force', '--output', str( pathlib.Path.cwd( ) / 'templates' ), '.', 'tests' )
        for x in glob.glob( str( pathlib.Path.cwd( ) / 'templates' / '*.rst' ) ) :
            with open( x, 'r' ) as fin :
                y = fin.read( ).replace( '   :members:', '   :members:\n   :noindex:' )
            with open( x, 'w' ) as fout :
                fout.write( y )
        for x in glob.glob( str( pathlib.Path.cwd( ) / 'templates' / 'modules.rst' ) ) :
            with open( x, 'r' ) as fin :
                y = fin.read( ).replace( 'noxfile', '' ).replace( 'setup', '' )
            with open( x, 'w' ) as fout :
                fout.write( y )
        for x in ( 'noxfile.rst', 'setup.rst' ) :
            if ( ( pathlib.Path.cwd( ) / 'templates' / x ).is_file( ) ) :
                os.remove( str( pathlib.Path.cwd( ) / 'templates' / x ) )
        session.run( 'sphinx-build', str( pathlib.Path.cwd( ) / 'templates' ), str( pathlib.Path.cwd( ) / 'docs' ) )
        session.run( 'git', 'add', str( pathlib.Path.cwd( ) / 'docs' / '*' ) )
        session.run( 'git', 'add', str( pathlib.Path.cwd( ) / 'templates' / '*' ) )

@nox.session( venv_backend = 'none' )
def image( session ) -> None :

    """ Build image.
    """

    if ( pathlib.Path( 'dockerfile' ).is_file( ) ) :
        build( session )
        login( session )
        session.run( 'docker', 'build', '--tag', repository, '--build-arg', 'JFROG_USER', '--build-arg', 'JFROG_PASSWD', '.' )

@nox.session( venv_backend = 'none' )
def login( session ) -> None :

    """ Login.
    """

    if ( pathlib.Path( 'dockerfile' ).is_file( ) ) :
        try :
            session.run( 'docker', 'login', 'global-artifacts.se.com', '-u', os.getenv( 'JFROG_USER' ), '-p', os.getenv( 'JFROG_PASSWD' ), external = True  )
        except Exception :
            pass

@nox.session( venv_backend = 'none' )
def notebook( session ) -> None :

    """ Run notebook.
    """

    if ( pathlib.Path( 'notebooks' ).is_dir( ) ) :
        os.chdir( 'notebooks' )
        value = [ x for x in glob.glob( '*.ipynb', recursive = True ) ]
        if ( value ) :
            session.run( 'jupyter', 'notebook', value[ 0 ] )

@nox.session( venv_backend = 'none' )
def push( session ) -> None :

    """ Push repository.
    """

    if ( pathlib.Path( '.git' ).is_dir( ) ) :
        package = repository.split( '-' )
        package = package[ max( len( package ) - 2, 0 ) ]
        if ( pathlib.Path( package ).is_dir( ) ) :
            session.run( 'git', 'add', str( pathlib.Path.cwd( ) / package / '*' ) )
        if ( pathlib.Path( 'service' ).is_dir( ) ) :
            session.run( 'git', 'add', str( pathlib.Path.cwd( ) / 'service' / '*' ) )
        if ( pathlib.Path( 'tests' ).is_dir( ) ) :
            session.run( 'git', 'add', str( pathlib.Path.cwd( ) / 'tests' / '*' ) )
        status( session )
        value = input( '[ ' + repository + ' ] message : ' )
        if ( value ) :
            if ( session.run( 'git', 'commit', '--all', '--message', value ) ) :
                session.run( 'git', 'push', 'origin', 'master' )
        try :
            url = 'https://github.schneider-electric.com'
            requests.request( method = 'head', url = url, timeout = 2 )
            value = input( '[ ' + repository + ' ] mirror : ' )
            if ( value ) :
                session.run( 'git', 'push', '--mirror', url + '/' + value + '/' + repository + '.git' )                
        except Exception :
            pass

@nox.session( venv_backend = 'none' )
def status( session ) -> None :

    """ Check status.
    """

    if ( pathlib.Path( '.git' ).is_dir( ) ) :
        print( '[ ' + repository + ' ]' )
        session.run( 'git', 'status', '--short' )

@nox.session( venv_backend = 'none' )
def tag( session ) -> None :

    """ Push tag.
    """

    if ( pathlib.Path( '.git' ).is_dir( ) ) :
        session.run( 'git', 'tag', '--list' )
        value = input( '[ ' + repository + ' ] annotate : ' )
        if ( value ) :
            session.run( 'git', 'tag', '--annotate', value, '--force', '--message', '.' )
            session.run( 'git', 'push', '--force', '--tags' )

@nox.session( venv_backend = 'none' )
def tests( session ) -> None :

    """ Run tests.
    """

    if ( pathlib.Path( 'tests' ).is_dir( ) ) :
        if ( [ u for u in pathlib.Path( 'tests' ).iterdir( ) ] ) :
            try :
                if ( pathlib.Path( 'docker-compose.yml' ).is_file( ) ) :
                    login( session )
                    session.run( 'docker', 'compose', 'up', '--detach' )
                    time.sleep( 10.0 )
                session.run( 'python', '-m', 'pip', 'install', '-e', '.' )
                session.run( 'pytest', '--capture=no', '--verbose', '-s' )
                shutil.rmtree( '.pytest_cache', ignore_errors = True )
            finally :
                if ( pathlib.Path( 'docker-compose.yml' ).is_file( ) ) :
                    session.run( 'docker', 'compose', 'down' )



""" **Description**
        A complex frequency filter adapts and discriminates the phase of a forward
        complex coefficient to produce a reference signal, which estimates a normalized
        frequency of a primary signal which is normalized to unity magnitude.  A
        normalized frequency and rate of adaptation are specified.

        .. math::
            f_{n} = \\frac{\\tan^{-1}(\ b_{n}\ ) }{\pi}

        .. math::
            x_{n} = \\frac{d_{n}}{|\ d_{n}\ |}

        .. math::
            y_{n} = b_{n} x_{n-1}

        .. math::
            e_{n} = d_{n} - y_{n}

        .. math::
            b_{0} = e^{\ j\ \pi\ f_{0}}

        .. math::
            b_{n} = b_{n} + \mu e_{n} x_{n}^{*}

    **Example**
       
        ::
        
            from diamondback import ComplexExponentialFilter
            import numpy

            x = numpy.linspace( 0.0, 0.1, 128 )

            # Create a primary signal.

            d = ComplexExponentialFilter( 0.0 ).filter( x )

            # Create an instance.

            obj = ComplexFrequencyFilter( frequency = 0.0, rate = 0.1 )

            # Filter a primary signal.

            obj.reset( d[ 0 ] )
            y, e, b = obj.filter( d )

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-02-01.
"""

from diamondback.filters.FirFilter import FirFilter
from typing import List, Tuple, Union
import math
import numpy
import scipy

class ComplexFrequencyFilter( FirFilter ) :

    """ Complex frequency filter.
    """

    @property
    def frequency( self ) :

        """ frequency : float - frequency normalized to Nyquist in [ -1.0, 1.0 ].
        """

        return self._frequency

    @frequency.setter
    def frequency( self, frequency : float ) :

        if ( ( frequency < -1.0 ) or ( frequency > 1.0 ) ) :
            raise ValueError( f'Frequency = {frequency} Expected Frequency in [ -1.0, 1.0 ]' )
        self.b[ 0 ] = numpy.exp( 1j * math.pi * frequency )
        self._frequency = frequency

    @property
    def rate( self ) :

        """ rate : float - in [ 0.0, 1.0 ].
        """

        return self._rate

    @rate.setter
    def rate( self, rate : float ) :

        if ( ( rate < 0.0 ) or ( rate > 1.0 ) ) :
            raise ValueError( f'Rate = {rate} Expected Rate in [ 0.0, 1.0 ]' )
        self._rate = rate

    def __init__( self, frequency : float, rate : float ) -> None :

        """ Initialize.

            Arguments :
                frequency : float - frequency normalized to Nyquist in [ -1.0, 1.0 ).
                rate : float - in [ 0.0, 1.0 ].
        """

        if ( ( frequency < -1.0 ) or ( frequency > 1.0 ) ) :
            raise ValueError( f'Frequency = {frequency} Expected Frequency in [ -1.0, 1.0 ]' )
        if ( ( rate < 0.0 ) or ( rate > 1.0 ) ) :
            raise ValueError( f'Rate = {rate} Expected Rate in [ 0.0, 1.0 ]' )
        super( ).__init__( b = numpy.ones( 1, complex ), s = numpy.ones( 1, complex ) )
        self._frequency, self._rate = frequency, rate

    def filter( self, d : Union[ List, numpy.ndarray ] ) -> Tuple[ numpy.ndarray, numpy.ndarray, numpy.ndarray ] :

        """ Filters a primary signal and produces a reference signal.

            Signals are Hilbert transformed to complex as necessary.

            Arguments :
                d : Union[ List, numpy.ndarray ] - primary signal.

            Returns :
                y : numpy.ndarray - reference signal.
                e : numpy.ndarray - error signal.
                b : numpy.ndarray - forward coefficient.
        """

        if ( ( not numpy.isscalar( d ) ) and ( not isinstance( d, numpy.ndarray ) ) ) :
            d = numpy.array( list( d ) )
        if ( not len( d ) ) :
            raise ValueError( f'D = {d}' )
        if ( not numpy.iscomplex( d ).any( ) ) :
            d = scipy.signal.hilbert( d )
        x = abs( d )
        x[ numpy.isclose( x, 0.0 ) ] = 1.0
        x = d / x
        y, e, b = numpy.zeros( len( x ) ), numpy.zeros( len( x ), complex ), numpy.zeros( len( x ), complex )
        for ii in range( 0, len( x ) ) :
            y[ ii ] = numpy.angle( self.b[ 0 ] ) / math.pi
            e[ ii ] = x[ ii ] - self.b[ 0 ] * self.s[ 0 ]
            b[ ii ] = self.b[ 0 ]
            self.b[ 0 ] += self.rate * e[ ii ] * numpy.conjugate( self.s[ 0 ] )
            self.s[ 0 ] = x[ ii ]
        return y, e, b

    def reset( self, x : complex ) -> None :

        """ Modifies a state to minimize edge effects by assuming persistent
            operation at a specified primary incident condition.

            Arguments :
                x : complex - incident signal.
        """

        if ( not numpy.isscalar( x ) ) :
            raise ValueError( f'X = {x}' )
        if ( numpy.isclose( x, 0.0 ) ) :
            self.s[ 0 ] = 1.0
        else :
            self.s[ 0 ] = x / abs( x )



""" **Description**
        A Gaussian model is a supervised learning probabilistic model instance
        which uses maximum likelihood estimation and regularization to maximize
        posterior probability and classify an incident signal.  Learns one
        distribution instance per class.

    **Example**
      
        ::
        
            from diamondback import GaussianModel

            # Create an instance.

            obj = GaussianModel( )

            # Learn an incident signal and predict classification.

            x, y = numpy.random.rand( 32, 2 ), numpy.random.randint( 0, 10, 32 )
            obj.learn( x, y )
            x = numpy.random.rand( 16, 2 )
            v = obj.predict( x )

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-02-08.
"""

import numpy

class GaussianModel( object ) :

    """ Gaussian model.
    """

    @property
    def regularize( self ) :
        return self._regularize

    @regularize.setter
    def regularize( self, regularize : float ) :
        if ( regularize < 0.0 ) :
            raise ValueError( f'Regularize = {regularize} Expected Regularize in [ 0.0, inf )' )
        self._regularize = regularize

    @property
    def shape( self ) :
        return self._shape
        
    def __init__( self, regularize : float = 1.0e-1 ) -> None :

        """ Initialize.

            Arguments :
                regularize : float - regularize.
        """

        if ( regularize < 0.0 ) :
            raise ValueError( f'Regularize = {regularize} Expected Regularize in [ 0.0, inf )' )
        self._data = [ ]
        self._regularize = regularize
        self._shape = ( )

    def learn( self, x : numpy.ndarray, y : numpy.ndarray ) -> None :

        """ Learns an incident signal with ground truth label and estimates inverse
            covariance and mean matrices to learn a distribution instance for
            each class.
        
            Arguments :
                x : numpy.ndarray ( batch, count ) - incident.
                y : numpy.ndarray ( batch ) - label.
        """

        if ( ( len( x.shape ) != 2 ) or ( not all( x.shape ) ) ) :
            raise ValueError( f'X = {x.shape}' )
        if ( len( y ) != x.shape[ 0 ] ) :
            raise ValueError( f'Y = {len( y )} Expected Y = {x.shape[ 0 ]}' )
        if ( not issubclass( y.dtype.type, numpy.integer ) ) :
            raise ValueError( f'Y = {y.dtype.type} Expected Y = {numpy.integer}' )
        self._data = [ ]
        self._shape = x[ 0 ].shape
        r = self.regularize * numpy.identity( x.shape[ 1 ] )
        for ii in sorted( set( y ) ) :
            z = x[ numpy.where( y == ii )[ 0 ] ]
            u = z.mean( 0 )
            i = numpy.linalg.inv( ( ( z - u ).T @ ( z - u ) / z.shape[ 1 ] ) + r )
            self._data.append( dict( covariancei = i, mean = u ) )

    def predict( self, x : numpy.ndarray ) -> numpy.ndarray :

        """ Predicts an estimate of ground truth label from an incident signal
            and maximizes posterior probability.
            
            Predictions for each class are ranked and ordered by decending
            probability, and the initial prediction is the most likely class.
        
            Arguments :
                x : numpy.ndarray ( batch, count ) - data.

            Returns :
                v : numpy.ndarray ( batch, class ) - predict.
        """

        if ( not len( self._data ) ) :
            raise ValueError( f'Data = {self._data}' )
        if ( ( len( x.shape ) != 2 ) or ( not all( x.shape ) ) ) :
            raise ValueError( f'X = {x.shape}' )
        if ( x[ 0 ].shape != self._shape ) :
            raise ValueError( f'X = {x[ 0 ].shape} Expected X = {self._shape}' )
        if ( not len( self._data ) ) :
            raise RuntimeError( f'Model = {self._data} Not Trained' )
        v = numpy.zeros( ( x.shape[ 0 ], len( self._data ) ) )
        for jj in range( 0, len( v ) ) :
            for ii in range( 0, len( self._data ) ) :
                m = self._data[ ii ]
                i, u = m[ 'covariancei' ], m[ 'mean' ]
                v[ jj, ii ] = max( ( x[ jj ] - u ) @ i @ ( x[ jj ] - u ).T, 0.0 )
        return numpy.argsort( v, axis = 1 )



""" **Description**
        A complex exponential filter produces a complex exponential reference
        signal from an incident signal equal to a specified normalized frequency.
        A normalized phase is specified.

        .. math::
            x_{n} = f_{n}

        .. math::
            \phi_{n} = \phi_{n-1} + x_{n}

        .. math::
            y_{n} = e^{\ j\ \pi\ \phi_{n}}

    **Example**
      
        ::
        
            from diamondback import ComplexExponentialFilter
            import numpy

            # Create an instance.

            obj = ComplexExponentialFilter( phase = 0.0 )

            # Filter an incident signal.

            x = numpy.linspace( -1.0e-4, 1.0e-4, 128 ) + 0.1
            y = obj.filter( x )

    **License**
        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_
        © 2018 - 2023 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**
        Larry Turner, Schneider Electric, AI Hub, 2018-01-31.
"""

from typing import List, Union
import math
import numpy

class ComplexExponentialFilter( object ) :

    """ Complex exponential filter.
    """

    @property
    def phase( self ) :

        """ phase : float - relative to pi in [ -1.0, 1.0 ].
        """

        return self._phase

    @phase.setter
    def phase( self, phase : float ) :

        if ( ( phase < -1.0 ) or ( phase > 1.0 ) ) :
            raise ValueError( f'Phase = {phase} Expected Phase in [ -1.0, 1.0 ]' )
        self._phase = phase

    def __init__( self, phase : float = 0.0 ) -> None :

        """ Initialize.

            Arguments :
                phase : float - relative to pi in [ -1.0, 1.0 ].
        """

        if ( ( phase < -1.0 ) or ( phase > 1.0 ) ) :
            raise ValueError( f'Phase = {phase} Expected Phase in [ -1.0, 1.0 ]' )
        super( ).__init__( )
        self._phase = phase

    def filter( self, x : Union[ List, numpy.ndarray ] ) -> numpy.ndarray :

        """ Filters an incident signal and produces a reference signal.

            Arguments :
                x : Union[ List, numpy.ndarray ] - incident signal frequency normalized to Nyquist in [ -1.0, 1.0 ).

            Returns :
                y : numpy.ndarray - reference signal.
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :
            x = numpy.array( list( x ) )
        if ( ( numpy.iscomplex( x ).any( ) ) or ( not len( x ) ) or ( abs( x ) > 1.0 ).any( ) ) :
            raise ValueError( f'X = {x}' )
        y = numpy.zeros( len( x ), complex )
        phase = self.phase
        for ii in range( 0, len( x ) ) :
            y[ ii ] = numpy.exp( 1j * math.pi * phase )
            phase += x[ ii ]
        phase = numpy.fmod( phase, 2.0 )
        if ( abs( phase ) > 1.0 ) :
            phase -= numpy.sign( phase ) * 2.0
        self.phase = phase
        return y
